"""Provider for code formatting and linting tools."""

from __future__ import annotations

from llmling_agent.resource_providers.static import StaticResourceProvider
from llmling_agent.tools.base import Tool


async def format_code(code: str, language: str | None = None) -> str:
    """Format and lint code, returning a concise summary.

    Args:
        code: Source code to format and lint
        language: Programming language (auto-detected if not provided)

    Returns:
        Short status message about formatting/linting results
    """
    from anyenv.language_formatters import FormatterRegistry

    registry = FormatterRegistry("subprocess")
    registry.register_default_formatters()

    # Get formatter by language or try to detect
    formatter = None
    if language:
        formatter = registry.get_formatter_by_language(language)

    if not formatter:
        # Try to detect from content
        detected = registry.detect_language_from_content(code)
        if detected:
            formatter = registry.get_formatter_by_language(detected)

    if not formatter:
        return f"❌ Unsupported language: {language or 'unknown'}"

    try:
        result = await formatter.format_and_lint_string(code, fix=True)

        if result.success:
            changes = "formatted" if result.format_result.formatted else "no changes"
            lint_status = "clean" if result.lint_result.success else "has issues"
            duration = f"{result.total_duration:.2f}s"
            return f"✅ Code {changes}, {lint_status} ({duration})"
        errors = []
        if not result.format_result.success:
            errors.append(f"format: {result.format_result.error_type}")
        if not result.lint_result.success:
            errors.append(f"lint: {result.lint_result.error_type}")
        return f"❌ Failed: {', '.join(errors)}"

    except Exception as e:  # noqa: BLE001
        return f"❌ Error: {type(e).__name__}"


def create_code_tools() -> list[Tool]:
    """Create tools for code formatting and linting."""
    return [
        Tool.from_callable(format_code, source="builtin", category="execute"),
    ]


class CodeTools(StaticResourceProvider):
    """Provider for code formatting and linting tools."""

    def __init__(self, name: str = "code"):
        super().__init__(name=name, tools=create_code_tools())



"""Entry point based toolset implementation."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from epregistry import EntryPointRegistry

from llmling_agent.log import get_logger
from llmling_agent.resource_providers.base import ResourceProvider
from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


class EntryPointTools(ResourceProvider):
    """Provider for entry point based tools."""

    def __init__(self, module: str):
        super().__init__(name=module)
        self.module = module
        self._tools: list[Tool] | None = None
        self.registry = EntryPointRegistry[Callable[..., Any]]("llmling")

    async def get_tools(self) -> list[Tool]:
        """Get tools from entry points."""
        # Return cached tools if available
        if self._tools is not None:
            return self._tools

        self._tools = []
        entry_point = self.registry.get("tools")
        if not entry_point:
            msg = f"No tools entry point found for {self.module}"
            raise ValueError(msg)

        get_tools = entry_point.load()
        for item in get_tools():
            try:
                meta = {"module": self.module}
                tool = Tool.from_callable(item, source="entry_point", metadata=meta)
                self._tools.append(tool)
            except Exception as e:  # noqa: BLE001
                logger.warning("Failed to load tool", module=self.module, error=e)

        return self._tools



"""Event sources for LLMling agent."""

from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Any, Literal, Self

from pydantic import ConfigDict, Field
from schemez import Schema

from llmling_agent.messaging.messages import ChatMessage
from llmling_agent.talk.talk import Talk
from llmling_agent.utils.now import get_now
from llmling_agent_config.events import (
    ConnectionEventType,
    EventSourceConfig,  # noqa: TC001
)


ChangeType = Literal["added", "modified", "deleted"]


class EventData(Schema):
    """Base class for event data."""

    source: str
    timestamp: datetime = Field(default_factory=get_now)
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    @classmethod
    def create(cls, source: str, **kwargs: Any) -> Self:
        """Create event with current timestamp."""
        return cls(source=source, **kwargs)

    @abstractmethod
    def to_prompt(self) -> str:
        """Convert event to agent prompt."""

    async def format(self, config: EventSourceConfig) -> str:
        """Wraps core message with configurable template."""
        from jinjarope import Environment

        env = Environment(trim_blocks=True, lstrip_blocks=True, enable_async=True)
        template = env.from_string(config.template)

        return await template.render_async(
            source=self.source,
            content=self.to_prompt(),  # Use the core message
            metadata=self.metadata,
            timestamp=self.timestamp,
            include_metadata=config.include_metadata,
            include_timestamp=config.include_timestamp,
        )


class ConnectionEventData[TTransmittedData](EventData):
    """Event from connection activity."""

    connection_name: str
    """Name of the connection which fired an event."""

    connection: Talk[TTransmittedData]
    """The connection which fired the event."""

    event_type: ConnectionEventType
    """Type of event that occurred."""

    message: ChatMessage[TTransmittedData] | None = None
    """The message at the stage of the event."""

    def to_prompt(self) -> str:
        """Convert event to agent prompt."""
        base = f"Connection {self.connection_name!r} event: {self.event_type}"
        if self.message:
            return f"{base}\nMessage content: {self.message.content}"
        return base


class FileEventData(EventData):
    """File system event."""

    path: str
    type: ChangeType

    def to_prompt(self) -> str:
        return f"File {self.type}: {self.path}"


class FunctionResultEventData(EventData):
    """Event from a function execution result."""

    result: Any

    def to_prompt(self) -> str:
        """Convert result to prompt format."""
        return str(self.result)


class EmailEventData(EventData):
    """Email event with specific content structure."""

    subject: str
    sender: str
    body: str

    def to_prompt(self) -> str:
        """Core email message."""
        return f"Email from {self.sender} with subject: {self.subject}\n\n{self.body}"


class TimeEventData(EventData):
    """Time-based event."""

    schedule: str
    """Cron expression that triggered this event."""

    prompt: str
    """Cron expression that triggered this event."""

    def to_prompt(self) -> str:
        """Format scheduled event."""
        return f"Scheduled task triggered by {self.schedule}: {self.prompt}"


class WebhookEventData(EventData):
    """Webhook payload with formatting."""

    payload: dict[str, Any]

    def to_prompt(self) -> str:
        """Format webhook payload."""
        import anyenv

        return f"Webhook received:\n{anyenv.dump_json(self.payload, indent=True)}"



"""Manages message flow between agents/groups."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Sequence
from contextlib import asynccontextmanager
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any, Self, overload

from psygnal import Signal

from llmling_agent.log import get_logger
from llmling_agent.messaging.messages import ChatMessage
from llmling_agent.talk.stats import AggregatedTalkStats, TalkStats
from llmling_agent.utils.inspection import execute
from llmling_agent.utils.now import get_now


if TYPE_CHECKING:
    from collections.abc import Awaitable, Iterator
    from datetime import datetime, timedelta

    from llmling_agent.common_types import (
        AnyFilterFn,
        AnyTransformFn,
        ProcessorCallback,
        PromptCompatible,
        QueueStrategy,
    )
    from llmling_agent.messaging.events import ConnectionEventData, EventData
    from llmling_agent.messaging.messageemitter import MessageEmitter
    from llmling_agent.messaging.messagenode import MessageNode
    from llmling_agent_config.events import ConnectionEventType
    from llmling_agent_config.forward_targets import ConnectionType

logger = get_logger(__name__)


class Talk[TTransmittedData]:
    """Manages message flow between agents/groups."""

    @dataclass(frozen=True)
    class ConnectionProcessed:
        """Event emitted when a message flows through a connection."""

        message: ChatMessage
        source: MessageEmitter
        targets: list[MessageNode]
        queued: bool
        connection_type: ConnectionType
        timestamp: datetime = field(default_factory=get_now)

    # Original message "coming in"
    message_received = Signal(ChatMessage)
    # After any transformation (one for each message, not per target)
    message_forwarded = Signal(ChatMessage)
    # Comprehensive signal capturing all information about one "message handling process"
    connection_processed = Signal(ConnectionProcessed)

    def __init__(
        self,
        source: MessageEmitter,
        targets: Sequence[MessageNode],
        group: TeamTalk | None = None,
        *,
        name: str | None = None,
        connection_type: ConnectionType = "run",
        wait_for_connections: bool = False,
        priority: int = 0,
        delay: timedelta | None = None,
        queued: bool = False,
        queue_strategy: QueueStrategy = "latest",
        transform: AnyTransformFn[ChatMessage[TTransmittedData]] | None = None,
        filter_condition: AnyFilterFn | None = None,
        stop_condition: AnyFilterFn | None = None,
        exit_condition: AnyFilterFn | None = None,
    ):
        """Initialize talk connection.

        Args:
            source: Agent sending messages
            targets: Agents receiving messages
            group: Optional group this talk belongs to
            name: Optional name for this talk
            connection_type: How to handle messages:
                - "run": Execute message as a new run in target
                - "context": Add message as context to target
                - "forward": Forward message to target's outbox
            wait_for_connections: Whether to wait for all targets to complete
            priority: Task priority (lower = higher priority)
            delay: Optional delay before processing
            queued: Whether messages should be queued for manual processing
            queue_strategy: How to process queued messages:
                - "concat": Combine all messages with newlines
                - "latest": Use only the most recent message
                - "buffer": Process all messages individually
            transform: Optional function to transform messages
            filter_condition: Optional condition for filtering messages
            stop_condition: Optional condition for disconnecting
            exit_condition: Optional condition for stopping the event loop
        """
        self.source = source
        self.targets = list(targets)
        # Could perhaps better be an auto-inferring property
        self.name = name or f"{source.name}->{[t.name for t in targets]}"
        self.group = group
        self.priority = priority
        self.delay = delay
        self.active = True
        self.connection_type = connection_type
        self.wait_for_connections = wait_for_connections
        self.queued = queued
        self.queue_strategy = queue_strategy
        self._pending_messages = defaultdict[str, list[ChatMessage[TTransmittedData]]](
            list
        )
        names = {t.name for t in targets}
        self._stats = TalkStats(source_name=source.name, target_names=names)
        self.transform_fn = transform
        self.filter_condition = filter_condition
        self.stop_condition = stop_condition
        self.exit_condition = exit_condition

    def __repr__(self):
        targets = [t.name for t in self.targets]
        return f"<Talk({self.connection_type}) {self.source.name} -> {targets}>"

    @overload
    def __rshift__(
        self,
        other: MessageNode[Any, str]
        | ProcessorCallback[str]
        | Sequence[MessageNode[Any, str] | ProcessorCallback[str]],
    ) -> TeamTalk[str]: ...

    @overload
    def __rshift__(
        self,
        other: MessageNode[Any, Any]
        | ProcessorCallback[Any]
        | Sequence[MessageNode[Any, Any] | ProcessorCallback[Any]],
    ) -> TeamTalk[Any]: ...

    def __rshift__(
        self,
        other: MessageNode[Any, Any]
        | ProcessorCallback[Any]
        | Sequence[MessageNode[Any, Any] | ProcessorCallback[Any]],
    ) -> TeamTalk[Any]:
        """Add another node as target to the connection or group.

        Example:
            connection >> other_agent  # Connect to single agent
            connection >> (agent2 & agent3)  # Connect to group
        """
        from llmling_agent import Agent, MessageNode
        from llmling_agent.talk import TeamTalk

        match other:
            case Callable():
                other = Agent.from_callback(other)
                if pool := self.source.context.pool:
                    pool.register(other.name, other)
                return self.__rshift__(other)
            case Sequence():
                team_talks = [self.__rshift__(o) for o in other]
                return TeamTalk([self, *team_talks])
            case MessageNode():
                talks = [t.__rshift__(other) for t in self.targets]
                return TeamTalk([self, *talks])
            case _:
                msg = f"Invalid agent type: {type(other)}"
                raise TypeError(msg)

    async def _evaluate_condition(
        self,
        condition: Callable[..., bool | Awaitable[bool]] | None,
        message: ChatMessage[Any],
        target: MessageNode,
        *,
        default_return: bool = False,
    ) -> bool:
        """Evaluate a condition with flexible parameter handling."""
        from llmling_agent.talk.registry import EventContext

        if not condition:
            return default_return
        registry = (
            context.pool.connection_registry
            if (context := self.source.context) and context.pool
            else None
        )
        ctx = EventContext(
            message=message,
            target=target,
            stats=self.stats,
            registry=registry,
            talk=self,
        )
        return await execute(condition, ctx)

    def on_event(
        self,
        event_type: ConnectionEventType,
        callback: Callable[
            [ConnectionEventData[TTransmittedData]], None | Awaitable[None]
        ],
    ) -> Self:
        """Register callback for connection events."""
        from llmling_agent.messaging.events import ConnectionEventData

        async def wrapped_callback(event: EventData):
            if isinstance(event, ConnectionEventData) and event.event_type == event_type:
                await execute(callback, event)

        self.source._events.add_callback(wrapped_callback)
        return self

    async def _emit_connection_event(
        self,
        event_type: ConnectionEventType,
        message: ChatMessage[TTransmittedData] | None,
    ):
        from llmling_agent.messaging.events import ConnectionEventData

        event = ConnectionEventData[Any](
            connection=self,
            source="connection",
            connection_name=self.name,
            event_type=event_type,
            message=message,
            timestamp=get_now(),
        )
        # Propagate to all event managers through registry
        if self.source.context and (pool := self.source.context.pool):
            for connection in pool.connection_registry.values():
                await connection.source._events.emit_event(event)

    async def _handle_message(
        self,
        message: ChatMessage[TTransmittedData],
        prompt: str | None = None,
    ) -> list[ChatMessage[Any]]:
        """Handle message forwarding based on connection configuration."""
        # 2. Early exit checks
        if not (self.active and (not self.group or self.group.active)):
            return []

        # 3. Check exit condition for any target
        for target in self.targets:
            # Exit if condition returns True
            if await self._evaluate_condition(self.exit_condition, message, target):
                raise SystemExit

        # 4. Check stop condition for any target
        for target in self.targets:
            # Stop if condition returns True
            if await self._evaluate_condition(self.stop_condition, message, target):
                self.disconnect()
                return []

        # 5. Transform if configured
        processed_message = message
        if self.transform_fn:
            processed_message = await execute(self.transform_fn, message)
        # 6. First pass: Determine target list
        target_list: list[MessageNode] = [
            target
            for target in self.targets
            if await self._evaluate_condition(
                self.filter_condition,
                processed_message,
                target,
                default_return=True,
            )
        ]
        # 7. emit connection processed event
        self.connection_processed.emit(
            self.ConnectionProcessed(
                message=processed_message,
                source=self.source,
                targets=target_list,
                queued=self.queued,
                connection_type=self.connection_type,  # pyright: ignore
            )
        )
        # 8. if we have targets, update stats and emit message forwarded
        if target_list:
            messages = [*self._stats.messages, processed_message]
            self._stats = replace(self._stats, messages=messages)
            self.message_forwarded.emit(processed_message)

        # 9. Second pass: Actually process for each target
        responses: list[ChatMessage[Any]] = []
        for target in target_list:
            if self.queued:
                self._pending_messages[target.name].append(processed_message)
                continue
            if response := await self._process_for_target(
                processed_message, target, prompt
            ):
                responses.append(response)

        return responses

    async def _process_for_target(
        self,
        message: ChatMessage[Any],
        target: MessageNode,
        prompt: PromptCompatible | None = None,
    ) -> ChatMessage[Any] | None:
        """Process message for a single target."""
        from llmling_agent.agent import Agent
        from llmling_agent.delegation.base_team import BaseTeam

        match self.connection_type:
            case "run":
                prompts: list[PromptCompatible] = [message]
                if prompt:
                    prompts.append(prompt)
                return await target.run(*prompts)

            case "context":
                meta = {
                    "type": "forwarded_message",
                    "role": message.role,
                    "model": message.model_name,
                    "cost_info": message.cost_info,
                    "timestamp": message.timestamp.isoformat(),
                    "prompt": prompt,
                }

                async def add_context():
                    match target:
                        case BaseTeam():
                            # Use distribute for teams
                            await target.distribute(str(message.content), metadata=meta)
                        case Agent():  # Agent case
                            # Use existing context message approach
                            target.conversation.add_context_message(
                                str(message.content),
                                source=message.name,
                                metadata=meta,
                            )

                if self.delay is not None or self.priority != 0:
                    coro = add_context()
                    target.task_manager.run_background(
                        coro, priority=self.priority, delay=self.delay
                    )
                else:
                    await add_context()
                return None

            case "forward":
                if self.delay is not None or self.priority != 0:

                    async def delayed_emit():
                        await target.connections.route_message(message)

                    coro = delayed_emit()
                    target.task_manager.run_background(
                        coro, priority=self.priority, delay=self.delay
                    )
                else:
                    await target.connections.route_message(message)
                return None

    async def trigger(
        self, prompt: PromptCompatible | None = None
    ) -> list[ChatMessage[TTransmittedData]]:
        """Process queued messages."""
        if not self._pending_messages:
            return []
        match self.queue_strategy:
            case "buffer":
                results: list[ChatMessage[TTransmittedData]] = []
                # Process each agent's queue
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    for msg in queue:
                        if resp := await self._process_for_target(msg, target, prompt):
                            results.append(resp)  # noqa: PERF401
                    queue.clear()
                return results

            case "latest":
                results = []
                # Get latest message for each agent
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    if queue:
                        latest = queue[-1]
                        if resp := await self._process_for_target(latest, target, prompt):
                            results.append(resp)
                        queue.clear()
                return results

            case "concat":
                results = []
                # Concat messages per agent
                for target in self.targets:
                    queue = self._pending_messages[target.name]
                    if not queue:
                        continue

                    base = queue[-1]
                    contents = [str(m.content) for m in queue]
                    meta = {
                        **base.metadata,
                        "merged_count": len(queue),
                        "queue_strategy": self.queue_strategy,
                    }
                    content = "\n\n".join(contents)
                    merged = replace(base, content=content, metadata=meta)  # type: ignore

                    if response := await self._process_for_target(merged, target, prompt):
                        results.append(response)
                    queue.clear()

                return results
            case _:
                msg = f"Invalid queue strategy: {self.queue_strategy}"
                raise ValueError(msg)

    def when(self, condition: AnyFilterFn) -> Self:
        """Add condition for message forwarding."""
        self.filter_condition = condition
        return self

    def transform[TNewData](
        self,
        transformer: Callable[
            [ChatMessage[TTransmittedData]],
            ChatMessage[TNewData] | Awaitable[ChatMessage[TNewData]],
        ],
        *,
        name: str | None = None,
        description: str | None = None,
    ) -> Talk[TNewData]:
        """Chain a new transformation after existing ones.

        Args:
            transformer: Function to transform messages
            name: Optional name for debugging
            description: Optional description

        Returns:
            New Talk instance with chained transformation

        Example:
            ```python
            talk = (agent1 >> agent2)
                .transform(parse_json)      # str -> dict
                .transform(extract_values)  # dict -> list
            ```
        """
        new_talk = Talk[TNewData](
            source=self.source,
            targets=self.targets,
            connection_type=self.connection_type,  # type: ignore
        )

        if self.transform_fn is not None:
            oldtransform_fn = self.transform_fn

            async def chainedtransform_fn(
                data: ChatMessage[TTransmittedData],
            ) -> ChatMessage[TNewData]:
                intermediate = await execute(oldtransform_fn, data)
                return await execute(transformer, intermediate)

            new_talk.transform_fn = chainedtransform_fn  # type: ignore
        else:
            new_talk.transform_fn = transformer  # type: ignore

        return new_talk

    @asynccontextmanager
    async def paused(self):
        """Temporarily set inactive."""
        previous = self.active
        self.active = False
        try:
            yield self
        finally:
            self.active = previous

    def disconnect(self):
        """Permanently disconnect the connection."""
        self.active = False

    @property
    def stats(self) -> TalkStats:
        """Get current connection statistics."""
        return self._stats


class TeamTalk[TTransmittedData](list["Talk | TeamTalk"]):
    """Group of connections with aggregate operations."""

    def __init__(
        self, talks: Sequence[Talk[TTransmittedData] | TeamTalk[TTransmittedData]]
    ):
        super().__init__(talks)
        self.filter_condition: AnyFilterFn | None = None
        self.active = True

    def __repr__(self):
        return f"TeamTalk({list(self)})"

    def __rshift__(
        self,
        other: MessageNode[Any, Any]
        | ProcessorCallback[Any]
        | Sequence[MessageNode[Any, Any] | ProcessorCallback[Any]],
    ) -> TeamTalk[Any]:
        """Add another node as target to the connection or group.

        Example:
            connection >> other_agent  # Connect to single agent
            connection >> (agent2 & agent3)  # Connect to group
        """
        from llmling_agent import Agent, MessageNode
        from llmling_agent.talk import TeamTalk

        match other:
            case Callable():
                other = Agent.from_callback(other)
                for talk_ in self.iter_talks():
                    if pool := talk_.source.context.pool:
                        pool.register(other.name, other)
                        break
                return self.__rshift__(other)
            case Sequence():
                team_talks = [self.__rshift__(o) for o in other]
                return TeamTalk([self, *team_talks])
            case MessageNode():
                talks = [t.connect_to(other) for t in self.targets]
                return TeamTalk([self, *talks])
            case _:
                msg = f"Invalid agent type: {type(other)}"
                raise TypeError(msg)

    @property
    def targets(self) -> list[MessageNode]:
        """Get all targets from all connections."""
        return [t for talk in self for t in talk.targets]

    def iter_talks(self) -> Iterator[Talk]:
        """Get all contained talks."""
        for t in self:
            match t:
                case Talk():
                    yield t
                case TeamTalk():
                    yield from t.iter_talks()

    async def _handle_message(self, message: ChatMessage[Any], prompt: str | None = None):
        for talk in self:
            await talk._handle_message(message, prompt)

    async def trigger(self, prompt: PromptCompatible | None = None) -> list[ChatMessage]:
        messages = []
        for talk in self:
            messages.extend(await talk.trigger(prompt))
        return messages

    @classmethod
    def from_nodes(
        cls,
        agents: Sequence[MessageNode],
        targets: list[MessageNode] | None = None,
    ) -> Self:
        """Create TeamTalk from a collection of agents."""
        return cls([Talk(agent, targets or []) for agent in agents])

    @asynccontextmanager
    async def paused(self):
        """Temporarily set inactive."""
        previous = self.active
        self.active = False
        try:
            yield self
        finally:
            self.active = previous

    def has_active_talks(self) -> bool:
        """Check if any contained talks are active."""
        return any(talk.active for talk in self)

    def get_active_talks(self) -> list[Talk | TeamTalk]:
        """Get list of currently active talks."""
        return [talk for talk in self if talk.active]

    @property
    def stats(self) -> AggregatedTalkStats:
        """Get aggregated statistics for all connections."""
        return AggregatedTalkStats(stats=[talk.stats for talk in self])

    def when(self, condition: AnyFilterFn) -> Self:
        """Add condition to all connections in group."""
        for talk in self:
            talk.when(condition)
        return self

    def disconnect(self):
        """Disconnect all connections in group."""
        for talk in self:
            talk.disconnect()



"""Command bridge for converting slashed commands to ACP format."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

from acp.schema import AvailableCommand
from llmling_agent.log import get_logger
from llmling_agent_acp.commands.acp_commands import ACPCommandContext
from llmling_agent_acp.commands.mcp_commands import PromptCommand


if TYPE_CHECKING:
    from collections.abc import Callable

    from slashed import CommandContext, CommandStore

    from llmling_agent.agent.context import AgentContext
    from llmling_agent.mcp_server.manager import Prompt
    from llmling_agent_acp.session import ACPSession

logger = get_logger(__name__)
SLASH_PATTERN = re.compile(r"^/([\w-]+)(?:\s+(.*))?$")
ACP_COMMANDS = {"list-sessions", "load-session", "save-session", "delete-session"}


class ACPCommandBridge:
    """Converts slashed commands to ACP AvailableCommand format."""

    def __init__(self, command_store: CommandStore) -> None:
        """Initialize with existing command store.

        Args:
            command_store: The slashed CommandStore containing available commands
        """
        self.command_store = command_store
        self._update_callbacks: list[Callable[[], None]] = []
        self._mcp_prompt_commands: dict[str, PromptCommand] = {}

    def get_acp_commands(self, context: AgentContext[Any]) -> list[AvailableCommand]:
        """Convert slashed commands to ACP format.

        Args:
            context: Optional agent context to filter commands

        Returns:
            List of ACP AvailableCommand objects
        """
        commands = [
            AvailableCommand.create(
                name=cmd.name,
                description=cmd.description,
                input_hint=cmd.usage,
            )
            for cmd in self.command_store.list_commands()
        ]
        commands.extend([
            cmd.to_available_command() for cmd in self._mcp_prompt_commands.values()
        ])
        return commands

    async def execute_slash_command(self, command_text: str, session: ACPSession) -> None:
        """Execute slash command and send results immediately as ACP notifications.

        Args:
            command_text: Full command text (including slash)
            session: ACP session context
        """
        if match := SLASH_PATTERN.match(command_text.strip()):
            command_name = match.group(1)
            args = match.group(2) or ""
            command_name, args = command_name, args.strip()
        else:
            logger.warning("Invalid slash command", command=command_text)
            return

        # Check if it's an MCP prompt command first
        if command_name in self._mcp_prompt_commands:
            mcp_cmd = self._mcp_prompt_commands[command_name]
            await mcp_cmd.execute(args, session)
            return
        if command_name in ACP_COMMANDS:
            # Use ACP context for ACP commands

            acp_ctx = ACPCommandContext(session)
            cmd_ctx: CommandContext = self.command_store.create_context(
                data=acp_ctx,
                output_writer=session.notifications.send_agent_text,
            )
        else:
            # Use regular agent context for other commands
            cmd_ctx = self.command_store.create_context(
                data=session.agent.context,
                output_writer=session.notifications.send_agent_text,
            )

        command_str = f"{command_name} {args}".strip()
        try:
            await self.command_store.execute_command(command_str, cmd_ctx)
        except Exception as e:
            logger.exception("Command execution failed")
            await session.notifications.send_agent_text(f"❌ Command error: {e}")

    def register_update_callback(self, callback: Callable[[], None]) -> None:
        """Register callback for command updates.

        Args:
            callback: Function to call when commands are updated
        """
        self._update_callbacks.append(callback)

    def add_mcp_prompt_commands(self, prompts: list[Prompt]) -> None:
        """Add prompts as slash commands.

        Args:
            prompts: List of Prompt instances from ToolManager
        """
        self._mcp_prompt_commands = {p.name: PromptCommand(p) for p in prompts}
        self._notify_command_update()  # Notify about command updates

    def _notify_command_update(self) -> None:
        """Notify all registered callbacks about command updates."""
        for callback in self._update_callbacks:
            try:
                callback()
            except Exception:
                logger.exception("Command update callback failed")



"""Provider for user interaction tools."""

from __future__ import annotations

from typing import Any

from pydantic_ai import RunContext

from llmling_agent.agent.context import AgentContext  # noqa: TC001
from llmling_agent.resource_providers.static import StaticResourceProvider
from llmling_agent.tools.base import Tool


async def ask_user(  # noqa: D417
    ctx: AgentContext,
    prompt: str,
    response_schema: dict[str, Any] | None = None,
) -> str:
    """Allow LLM to ask user a clarifying question during processing.

    This tool enables agents to ask users for additional information or clarification
    when needed to complete a task effectively.

    Args:
        prompt: Question to ask the user
        response_schema: Optional JSON schema for structured response (defaults to string)

    Returns:
        The user's response as a string
    """
    from mcp.types import ElicitRequestParams, ElicitResult, ErrorData

    if isinstance(ctx, RunContext):
        ctx = ctx.deps

    schema = response_schema or {"type": "string"}  # string schema if no none provided
    params = ElicitRequestParams(message=prompt, requestedSchema=schema)
    result = await ctx.handle_elicitation(params)

    match result:
        case ElicitResult(action="accept", content=content):
            return str(content)
        case ElicitResult(action="cancel"):
            return "User cancelled the request"
        case ElicitResult():
            return "User declined to answer"
        case ErrorData(message=message):
            return f"Error: {message}"
        case _:
            return "Unknown error occurred"


def create_user_interaction_tools() -> list[Tool]:
    """Create tools for user interaction operations."""
    return [Tool.from_callable(ask_user, source="builtin", category="other")]


class UserInteractionTools(StaticResourceProvider):
    """Provider for user interaction tools."""

    def __init__(self, name: str = "user_interaction"):
        super().__init__(name=name, tools=create_user_interaction_tools())



"""The main Agent. Can do all sort of crazy things."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass, field
from os import PathLike
import time
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    TypedDict,
    get_origin,
    get_type_hints,
    overload,
)
from uuid import uuid4

from anyenv import MultiEventHandler, method_spawner
from llmling import Config, RuntimeConfig, ToolError
from llmling_models import function_to_model, infer_model
import logfire
from psygnal import Signal
from pydantic import ValidationError
from pydantic._internal import _typing_extra
from pydantic_ai import (
    Agent as PydanticAgent,
    AgentRunResultEvent,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartStartEvent,
    RunContext,
    ToolCallPart,
    ToolReturnPart,
    models,
)
import pydantic_ai._function_schema
from pydantic_ai.models import Model
from pydantic_ai.tools import GenerateToolJsonSchema
from upath import UPath

from llmling_agent.agent.events import (
    RichAgentStreamEvent,
    StreamCompleteEvent,
    ToolCallCompleteEvent,
    ToolCallProgressEvent,
    create_queuing_progress_handler,
)
from llmling_agent.common_types import IndividualEventHandler
from llmling_agent.log import get_logger
from llmling_agent.messaging.messagenode import MessageNode
from llmling_agent.messaging.messages import ChatMessage, TokenCost
from llmling_agent.prompts.builtin_provider import RuntimePromptProvider
from llmling_agent.prompts.convert import convert_prompts
from llmling_agent.resource_providers.runtime import RuntimeResourceProvider
from llmling_agent.talk.stats import MessageStats
from llmling_agent.tasks import JobError
from llmling_agent.tools import SkillsRegistry, Tool, ToolManager
from llmling_agent.utils.inspection import call_with_context, has_argument_type
from llmling_agent.utils.now import get_now
from llmling_agent.utils.result_utils import to_type
from llmling_agent.utils.streams import merge_queue_into_iterator
from llmling_agent.utils.tasks import TaskManager
from llmling_agent_config.session import MemoryConfig


if TYPE_CHECKING:
    from collections.abc import AsyncIterable, AsyncIterator, Coroutine, Sequence
    from datetime import datetime
    from types import TracebackType

    from llmling.config.models import Resource
    from llmling.prompts import PromptType
    from pydantic_ai import AgentStreamEvent, UsageLimits
    from pydantic_ai.output import OutputSpec
    from toprompt import AnyPromptType
    from upath.types import JoinablePathLike

    from llmling_agent.agent import AgentContext
    from llmling_agent.agent.events import RichAgentStreamEvent
    from llmling_agent.common_types import (
        AgentName,
        EndStrategy,
        ModelType,
        ProcessorCallback,
        PromptCompatible,
        SessionIdType,
        ToolType,
    )
    from llmling_agent.delegation import Team, TeamRun
    from llmling_agent.resource_providers.base import ResourceProvider
    from llmling_agent_config.mcp_server import MCPServerConfig
    from llmling_agent_config.output_types import StructuredResponseConfig
    from llmling_agent_config.session import SessionQuery
    from llmling_agent_config.task import Job
    from llmling_agent_input.base import InputProvider


logger = get_logger(__name__)
# OutputDataT = TypeVar('OutputDataT', default=str, covariant=True)
NoneType = type(None)


class AgentKwargs(TypedDict, total=False):
    """Keyword arguments for configuring an Agent instance."""

    description: str | None
    model: ModelType
    system_prompt: str | Sequence[str]
    runtime: RuntimeConfig | Config | JoinablePathLike | None
    tools: Sequence[ToolType | Tool] | None
    toolsets: Sequence[ResourceProvider] | None
    mcp_servers: Sequence[str | MCPServerConfig] | None
    retries: int
    output_retries: int | None
    end_strategy: EndStrategy
    context: AgentContext[Any] | None  # x
    session: SessionIdType | SessionQuery | MemoryConfig | bool | int
    input_provider: InputProvider | None
    debug: bool
    event_handlers: Sequence[IndividualEventHandler] | None


class Agent[TDeps = None, OutputDataT = str](MessageNode[TDeps, OutputDataT]):
    """The main agent class.

    Generically typed with: LLMLingAgent[Type of Dependencies, Type of Result]
    """

    @dataclass(frozen=True)
    class AgentReset:
        """Emitted when agent is reset."""

        agent_name: AgentName
        previous_tools: dict[str, bool]
        new_tools: dict[str, bool]
        timestamp: datetime = field(default_factory=get_now)

    run_failed = Signal(str, Exception)
    agent_reset = Signal(AgentReset)

    def __init__(  # noqa: PLR0915
        # we dont use AgentKwargs here so that we can work with explicit ones in the ctor
        self,
        name: str = "llmling-agent",
        *,
        deps_type: type[TDeps] | None = None,
        model: ModelType = None,
        output_type: OutputSpec[OutputDataT] | StructuredResponseConfig | str = str,  # type: ignore[assignment]
        runtime: RuntimeConfig | Config | JoinablePathLike | None = None,
        context: AgentContext[TDeps] | None = None,
        session: SessionIdType | SessionQuery | MemoryConfig | bool | int = None,
        system_prompt: AnyPromptType | Sequence[AnyPromptType] = (),
        description: str | None = None,
        tools: Sequence[ToolType | Tool] | None = None,
        toolsets: Sequence[ResourceProvider] | None = None,
        mcp_servers: Sequence[str | MCPServerConfig] | None = None,
        resources: Sequence[Resource | PromptType | str] = (),
        retries: int = 1,
        output_retries: int | None = None,
        end_strategy: EndStrategy = "early",
        input_provider: InputProvider | None = None,
        parallel_init: bool = True,
        debug: bool = False,
        event_handlers: Sequence[IndividualEventHandler] | None = None,
    ):
        """Initialize agent with runtime configuration.

        Args:
            name: Name of the agent for logging and identification
            deps_type: Type of dependencies to use
            model: The default model to use (defaults to GPT-5)
            output_type: The default output type to use (defaults to str)
            runtime: Runtime configuration providing access to resources/tools
            context: Agent context with configuration
            session: Memory configuration.
                - None: Default memory config
                - False: Disable message history (max_messages=0)
                - int: Max tokens for memory
                - str/UUID: Session identifier
                - MemoryConfig: Full memory configuration
                - MemoryProvider: Custom memory provider
                - SessionQuery: Session query

            system_prompt: System prompts for the agent
            description: Description of the Agent ("what it can do")
            tools: List of tools to register with the agent
            toolsets: List of toolset resource providers for the agent
            mcp_servers: MCP servers to connect to
            resources: Additional resources to load
            retries: Default number of retries for failed operations
            output_retries: Max retries for result validation (defaults to retries)
            end_strategy: Strategy for handling tool calls that are requested alongside
                          a final result
            input_provider: Provider for human input (tool confirmation / HumanProviders)
            parallel_init: Whether to initialize resources in parallel
            debug: Whether to enable debug mode
            event_handlers: Sequence of event handlers to register with the agent
        """
        from llmling_agent.agent import AgentContext
        from llmling_agent.agent.conversation import ConversationManager
        from llmling_agent.agent.interactions import Interactions
        from llmling_agent.agent.sys_prompts import SystemPrompts

        self.task_manager = TaskManager()
        self._infinite = False
        # save some stuff for asnyc init
        self._owns_runtime = False
        self.deps_type = deps_type
        # match output_type:
        #     case type() | str():
        #         # For types and named definitions, use overrides if provided
        #         self.set_output_type(
        #             output_type,
        #             tool_name=tool_name,
        #             tool_description=tool_description,
        #         )
        #     case StructuredResponseConfig():
        #         # For response definitions, use as-is
        #         # (overrides don't apply to complete definitions)
        #         self.set_output_type(output_type)
        # prepare context
        ctx = context or AgentContext[TDeps].create_default(
            name,
            input_provider=input_provider,
        )
        self._context = ctx
        self._output_type = to_type(output_type, ctx.definition.responses)
        memory_cfg = (
            session
            if isinstance(session, MemoryConfig)
            else MemoryConfig.from_value(session)
        )
        # Initialize progress queue before super().__init__()
        self._progress_queue = asyncio.Queue[ToolCallProgressEvent]()

        super().__init__(
            name=name,
            context=ctx,
            description=description,
            enable_logging=memory_cfg.enable,
            mcp_servers=mcp_servers,
            progress_handler=create_queuing_progress_handler(self._progress_queue),
        )
        # Initialize runtime
        match runtime:
            case None:
                ctx.runtime = RuntimeConfig.from_config(Config())
            case Config() | str() | PathLike() | UPath():
                ctx.runtime = RuntimeConfig.from_config(runtime)
            case RuntimeConfig():
                ctx.runtime = runtime
            case _:
                msg = f"Invalid runtime type: {type(runtime)}"
                raise TypeError(msg)

        runtime_provider = RuntimePromptProvider(ctx.runtime)
        ctx.definition.prompt_manager.providers["runtime"] = runtime_provider
        # Initialize tool manager
        self.event_handler = MultiEventHandler[IndividualEventHandler](event_handlers)
        all_tools = list(tools or [])
        self.tools = ToolManager(all_tools)
        self.tools.add_provider(self.mcp)
        if builtin_tools := ctx.config.get_tool_provider():
            self.tools.add_provider(builtin_tools)

        for toolset_provider in toolsets or []:
            self.tools.add_provider(toolset_provider)

        # Initialize conversation manager
        resources = list(resources)
        if ctx.config.knowledge:
            resources.extend(ctx.config.knowledge.get_resources())
        self.conversation = ConversationManager(self, memory_cfg, resources=resources)

        # Store pydantic-ai configuration
        if model and not isinstance(model, str):
            assert isinstance(model, models.Model)
        self._model = model
        self._retries = retries
        self._end_strategy = end_strategy
        self._output_retries = output_retries

        self.skills_registry = SkillsRegistry()

        if ctx and ctx.definition:
            from llmling_agent.observability import registry

            registry.configure_observability(ctx.definition.observability)

        # init variables
        self._debug = debug
        self.parallel_init = parallel_init
        self._name = name
        self._background_task: asyncio.Task[Any] | None = None

        self.talk = Interactions(self)

        # Set up system prompts
        config_prompts = ctx.config.system_prompts if ctx else []
        all_prompts: list[AnyPromptType] = list(config_prompts)
        if isinstance(system_prompt, list):
            all_prompts.extend(system_prompt)
        else:
            all_prompts.append(system_prompt)
        self.sys_prompts = SystemPrompts(all_prompts, context=ctx)

    def __repr__(self) -> str:
        desc = f", {self.description!r}" if self.description else ""
        return f"Agent({self.name!r}, model={self._model!r}{desc})"

    def __prompt__(self) -> str:
        typ = self.__class__.__name__
        model = self.model_name or "default"
        parts = [f"Agent: {self.name}", f"Type: {typ}", f"Model: {model}"]
        if self.description:
            parts.append(f"Description: {self.description}")
        parts.extend([self.tools.__prompt__(), self.conversation.__prompt__()])
        return "\n".join(parts)

    async def __aenter__(self) -> Self:
        """Enter async context and set up MCP servers."""
        try:
            # Collect all coroutines that need to be run
            coros: list[Coroutine[Any, Any, Any]] = []

            # Runtime initialization if needed
            runtime_ref = self.context.runtime
            if runtime_ref and not runtime_ref._initialized:
                self._owns_runtime = True
                coros.append(runtime_ref.__aenter__())

            coros.append(super().__aenter__())
            coros.extend(self.conversation.get_initialization_tasks())

            # Execute coroutines either in parallel or sequentially
            if self.parallel_init and coros:
                await asyncio.gather(*coros)
            else:
                for coro in coros:
                    await coro
            if runtime_ref:
                self.tools.add_provider(RuntimeResourceProvider(runtime_ref))
            for provider in self.context.config.get_toolsets():
                self.tools.add_provider(provider)
        except Exception as e:
            # Clean up in reverse order
            if self._owns_runtime and runtime_ref and self.context.runtime == runtime_ref:
                await runtime_ref.__aexit__(type(e), e, e.__traceback__)
            msg = "Failed to initialize agent"
            raise RuntimeError(msg) from e
        else:
            return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        """Exit async context."""
        await super().__aexit__(exc_type, exc_val, exc_tb)
        try:
            await self.mcp.__aexit__(exc_type, exc_val, exc_tb)
        finally:
            if self._owns_runtime and self.context.runtime:
                self.tools.remove_provider("runtime")
                await self.context.runtime.__aexit__(exc_type, exc_val, exc_tb)
            # for provider in await self.context.config.get_toolsets():
            #     self.tools.remove_provider(provider.name)

    @overload
    def __and__(  # if other doesnt define deps, we take the agents one
        self, other: ProcessorCallback[Any] | Team[TDeps] | Agent[TDeps, Any]
    ) -> Team[TDeps]: ...

    @overload
    def __and__(  # otherwise, we dont know and deps is Any
        self, other: ProcessorCallback[Any] | Team[Any] | Agent[Any, Any]
    ) -> Team[Any]: ...

    def __and__(self, other: MessageNode[Any, Any] | ProcessorCallback[Any]) -> Team[Any]:
        """Create sequential team using & operator.

        Example:
            group = analyzer & planner & executor  # Create group of 3
            group = analyzer & existing_group  # Add to existing group
        """
        from llmling_agent.delegation.team import Team

        match other:
            case Team():
                return Team([self, *other.agents])
            case Callable():
                agent_2 = Agent.from_callback(other)
                agent_2.context.pool = self.context.pool
                return Team([self, agent_2])
            case MessageNode():
                return Team([self, other])
            case _:
                msg = f"Invalid agent type: {type(other)}"
                raise ValueError(msg)

    @overload
    def __or__(self, other: MessageNode[TDeps, Any]) -> TeamRun[TDeps, Any]: ...

    @overload
    def __or__[TOtherDeps](
        self,
        other: MessageNode[TOtherDeps, Any],
    ) -> TeamRun[Any, Any]: ...

    @overload
    def __or__(self, other: ProcessorCallback[Any]) -> TeamRun[Any, Any]: ...

    def __or__(self, other: MessageNode[Any, Any] | ProcessorCallback[Any]) -> TeamRun:
        # Create new execution with sequential mode (for piping)
        from llmling_agent import TeamRun

        if callable(other):
            other = Agent.from_callback(other)
            other.context.pool = self.context.pool

        return TeamRun([self, other])

    @classmethod
    def from_callback[TResult](
        cls,
        callback: ProcessorCallback[TResult],
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> Agent[None, TResult]:
        """Create an agent from a processing callback.

        Args:
            callback: Function to process messages. Can be:
                - sync or async
                - with or without context
                - must return str for pipeline compatibility
            name: Optional name for the agent
            kwargs: Additional arguments for agent
        """
        name = name or callback.__name__ or "processor"

        model = function_to_model(callback)
        # Get return type from signature for validation
        hints = get_type_hints(callback)
        return_type = hints.get("return")

        # If async, unwrap from Awaitable
        if (
            return_type
            and hasattr(return_type, "__origin__")
            and return_type.__origin__ is Awaitable
        ):
            return_type = return_type.__args__[0]
        return Agent(
            model=model,
            name=name,
            output_type=return_type or str,
            **kwargs,
        )

    @property
    def name(self) -> str:
        """Get agent name."""
        return self._name or "llmling-agent"

    @property
    def context(self) -> AgentContext[TDeps]:
        """Get agent context."""
        return self._context

    @context.setter
    def context(self, value: AgentContext[TDeps]):
        """Set agent context and propagate to provider."""
        self._context = value
        self.mcp.context = value
        self._context = value

    def set_output_type(
        self,
        output_type: type | str | StructuredResponseConfig | None,
        *,
        tool_name: str | None = None,
        tool_description: str | None = None,
    ):
        """Set or update the result type for this agent.

        Args:
            output_type: New result type, can be:
                - A Python type for validation
                - Name of a response definition
                - Response definition instance
                - None to reset to unstructured mode
            tool_name: Optional override for tool name
            tool_description: Optional override for tool description
        """
        self.log.debug("Setting result type", output_type=output_type)
        self._output_type = to_type(output_type)

    def to_structured[NewOutputDataT](
        self,
        output_type: type[NewOutputDataT] | str | StructuredResponseConfig,
        *,
        tool_name: str | None = None,
        tool_description: str | None = None,
    ) -> Agent[TDeps, NewOutputDataT] | Self:
        """Convert this agent to a structured agent.

        Args:
            output_type: Type for structured responses. Can be:
                - A Python type (Pydantic model)
                - Name of response definition from context
                - Complete response definition
            tool_name: Optional override for result tool name
            tool_description: Optional override for result tool description

        Returns:
            Typed Agent
        """
        self.set_output_type(output_type)  # type: ignore
        return self

    def is_busy(self) -> bool:
        """Check if agent is currently processing tasks."""
        return bool(self.task_manager._pending_tasks or self._background_task)

    @property
    def model_name(self) -> str | None:
        """Get the model name in a consistent format."""
        return self._model.model_name if isinstance(self._model, Model) else self._model

    def to_tool(
        self,
        *,
        name: str | None = None,
        reset_history_on_run: bool = True,
        pass_message_history: bool = False,
        parent: Agent[Any, Any] | None = None,
    ) -> Tool:
        """Create a tool from this agent.

        Args:
            name: Optional tool name override
            reset_history_on_run: Clear agent's history before each run
            pass_message_history: Pass parent's message history to agent
            parent: Optional parent agent for history/context sharing
        """
        tool_name = name or f"ask_{self.name}"

        # TODO: should probably make output type configurable
        async def wrapped_tool(prompt: str) -> Any:
            if pass_message_history and not parent:
                msg = "Parent agent required for message history sharing"
                raise ToolError(msg)

            if reset_history_on_run:
                self.conversation.clear()

            history = None
            if pass_message_history and parent:
                history = parent.conversation.get_history()
                old = self.conversation.get_history()
                self.conversation.set_history(history)
            result = await self.run(prompt)
            if history:
                self.conversation.set_history(old)
            return result.data

        normalized_name = self.name.replace("_", " ").title()
        docstring = f"Get expert answer from specialized agent: {normalized_name}"
        if self.description:
            docstring = f"{docstring}\n\n{self.description}"

        wrapped_tool.__doc__ = docstring
        wrapped_tool.__name__ = tool_name

        return Tool.from_callable(
            wrapped_tool,
            name_override=tool_name,
            description_override=docstring,
        )

    async def get_agentlet(
        self,
        tools: list[Tool],
        model: ModelType = None,
        output_type: type[Any] = str,
        deps_type: type[Any] | None = None,
    ):
        """Create pydantic-ai agent from current state."""
        # Monkey patch pydantic-ai to recognize AgentContext
        from llmling_agent.agent.context import AgentContext
        from llmling_agent.agent.tool_wrapping import wrap_tool

        def _is_call_ctx_patched(annotation):
            if annotation is RunContext or (
                _typing_extra.is_generic_alias(annotation)
                and get_origin(annotation) is RunContext
            ):
                return True
            return annotation is AgentContext or (
                _typing_extra.is_generic_alias(annotation)
                and get_origin(annotation) is AgentContext
            )

        original_is_call_ctx = pydantic_ai._function_schema._is_call_ctx
        pydantic_ai._function_schema._is_call_ctx = _is_call_ctx_patched

        actual_model = model or self._model
        model_ = (
            infer_model(actual_model) if isinstance(actual_model, str) else actual_model
        )

        agent = PydanticAgent(
            model=model_,
            instructions=await self.sys_prompts.format_system_prompt(self),
            retries=self._retries,
            end_strategy=self._end_strategy,
            output_retries=self._output_retries,
            deps_type=deps_type or NoneType,
            output_type=output_type,
        )

        for tool in tools:
            wrapped = wrap_tool(tool, self.context)
            if has_argument_type(wrapped, RunContext):
                agent.tool(wrapped)
            elif has_argument_type(wrapped, AgentContext):
                agent._function_toolset.add_function(
                    func=wrapped,
                    takes_ctx=True,
                    name=None,
                    retries=1,
                    prepare=None,
                    docstring_format="auto",
                    require_parameter_descriptions=False,
                    schema_generator=GenerateToolJsonSchema,
                    strict=None,
                )
            else:
                agent.tool_plain(wrapped)

        # Restore original function after agent creation
        pydantic_ai._function_schema._is_call_ctx = original_is_call_ctx

        return agent

    async def _run(
        self,
        *prompts: PromptCompatible | ChatMessage[Any],
        output_type: type[OutputDataT] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        tool_choice: str | list[str] | None = None,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        messages: list[ChatMessage[Any]] | None = None,
        deps: TDeps | None = None,
        wait_for_connections: bool | None = None,
    ) -> ChatMessage[OutputDataT]:
        """Run agent with prompt and get response.

        Args:
            prompts: User query or instruction
            output_type: Optional type for structured responses
            model: Optional model override
            store_history: Whether the message exchange should be added to the
                            context window
            tool_choice: Filter tool choice by name
            usage_limits: Optional usage limits for the model
            message_id: Optional message id for the returned message.
                        Automatically generated if not provided.
            conversation_id: Optional conversation id for the returned message.
            messages: Optional list of messages to replace the conversation history
            deps: Optional dependencies for the agent
            wait_for_connections: Whether to wait for connected agents to complete

        Returns:
            Result containing response and run information

        Raises:
            UnexpectedModelBehavior: If the model fails or behaves unexpectedly
        """
        """Run agent with prompt and get response."""
        message_id = message_id or str(uuid4())
        tools = await self.tools.get_tools(state="enabled", names=tool_choice)
        final_type = to_type(output_type) if output_type else self._output_type
        start_time = time.perf_counter()

        message_history = (
            messages if messages is not None else self.conversation.get_history()
        )
        try:
            # Create pydantic-ai agent for this run
            agentlet = await self.get_agentlet(tools, model, final_type, self.deps_type)
            converted_prompts = await convert_prompts(prompts)

            # Merge internal and external event handlers like the old provider did
            async def event_distributor(
                ctx: RunContext, events: AsyncIterable[AgentStreamEvent]
            ):
                async for event in events:
                    for handler in self.event_handler._wrapped_handlers:
                        await handler(ctx, event)

            result = await agentlet.run(
                [
                    i if isinstance(i, str) else i.to_pydantic_ai()
                    for i in converted_prompts
                ],
                deps=deps,
                message_history=[
                    m for run in message_history for m in run.to_pydantic_ai()
                ],
                output_type=final_type or str,
                usage_limits=usage_limits,
                event_stream_handler=event_distributor,
            )

            # Calculate costs
            usage = result.usage()
            cost_info = await TokenCost.from_usage(
                model=result.response.model_name, usage=usage
            )

            response_msg = ChatMessage[OutputDataT](
                content=result.output,
                role="assistant",
                name=self.name,
                model_name=result.response.model_name,
                finish_reason=result.response.finish_reason,
                messages=result.new_messages(),
                provider_response_id=result.response.provider_response_id,
                usage=usage,
                provider_name=result.response.provider_name,
                message_id=message_id,
                conversation_id=conversation_id,
                cost_info=cost_info,
                response_time=time.perf_counter() - start_time,
                provider_details={},
            )

        except Exception as e:
            self.log.exception("Agent run failed")
            self.run_failed.emit("Agent run failed", e)
            raise
        else:
            return response_msg

    @method_spawner
    async def run_stream(  # noqa: PLR0915
        self,
        *prompt: PromptCompatible,
        output_type: type[OutputDataT] | None = None,
        model: ModelType = None,
        tool_choice: str | list[str] | None = None,
        store_history: bool = True,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        messages: list[ChatMessage[Any]] | None = None,
        wait_for_connections: bool | None = None,
        deps: TDeps | None = None,
    ) -> AsyncIterator[RichAgentStreamEvent[OutputDataT]]:
        """Run agent with prompt and get a streaming response.

        Args:
            prompt: User query or instruction
            output_type: Optional type for structured responses
            model: Optional model override
            tool_choice: Filter tool choice by name
            store_history: Whether the message exchange should be added to the
                           context window
            usage_limits: Optional usage limits for the model
            message_id: Optional message id for the returned message.
                        Automatically generated if not provided.
            conversation_id: Optional conversation id for the returned message.
            messages: Optional list of messages to replace the conversation history
            wait_for_connections: Whether to wait for connected agents to complete
            deps: Optional dependencies for the agent
        Returns:
            An async iterator yielding streaming events with final message embedded.

        Raises:
            UnexpectedModelBehavior: If the model fails or behaves unexpectedly
        """
        message_id = message_id or str(uuid4())
        user_msg, prompts = await self.pre_run(*prompt)
        final_type = to_type(output_type) if output_type else self._output_type
        start_time = time.perf_counter()
        tools = await self.tools.get_tools(state="enabled", names=tool_choice)
        message_history = (
            messages if messages is not None else self.conversation.get_history()
        )
        try:
            agentlet = await self.get_agentlet(tools, model, final_type, self.deps_type)
            content = await convert_prompts(prompts)
            # Initialize variables for final response
            usage = None
            model_name = None
            new_msgs = []
            finish_reason = None
            provider_name = None
            provider_response_id = None
            output = None

            # Create tool dict for signal emission
            converted = [i if isinstance(i, str) else i.to_pydantic_ai() for i in content]
            stream_events = agentlet.run_stream_events(
                converted,
                deps=deps,
                message_history=[
                    m for run in message_history for m in run.to_pydantic_ai()
                ],
                output_type=final_type or str,
                usage_limits=usage_limits,
            )

            # Stream events through merge_queue for progress events
            async with merge_queue_into_iterator(
                stream_events, self._progress_queue
            ) as events:
                # Track tool call starts to combine with results later
                pending_tool_calls: dict[str, dict[str, Any]] = {}
                tool_dict = {i.name: i for i in tools}

                async for event in events:
                    # Process events and emit signals
                    match event:
                        case (
                            PartStartEvent(part=ToolCallPart() as tool_part)
                            | FunctionToolCallEvent(part=tool_part)
                        ) if tool_part.tool_name in tool_dict:
                            # Store tool call start info for later combination with result
                            pending_tool_calls[tool_part.tool_call_id] = {
                                "tool_name": tool_part.tool_name,
                                "tool_input": tool_part.args_as_dict(),
                            }
                            yield event
                        case (
                            FunctionToolResultEvent(tool_call_id=call_id) as result_event
                        ):
                            # Check if we have a pending tool call to combine with
                            if call_info := pending_tool_calls.pop(call_id, None):
                                # Create and yield combined event
                                combined_event = ToolCallCompleteEvent(
                                    tool_name=call_info["tool_name"],
                                    tool_call_id=call_id,
                                    tool_input=call_info["tool_input"],
                                    tool_result=result_event.result.content
                                    if isinstance(result_event.result, ToolReturnPart)
                                    else result_event.result,
                                    agent_name=self.name,
                                    message_id=message_id,
                                )
                                yield combined_event
                            yield event
                        case AgentRunResultEvent() as result_event:
                            # Capture final result data
                            result = result_event.result
                            usage = result.usage()
                            model_name = result.response.model_name
                            new_msgs = result.new_messages()
                            finish_reason = result.response.finish_reason
                            provider_name = result.response.provider_name
                            provider_response_id = result.response.provider_response_id
                            output = result.output
                            # Yield the original AgentRunResultEvent
                            yield event
                        case _:
                            # Yield all other events (including progress events)
                            yield event

            # Build final response message
            cost_info = None
            if model_name and usage and model_name != "test":
                cost_info = await TokenCost.from_usage(usage, model_name)

            response_msg = ChatMessage[OutputDataT](
                content=output,  # type: ignore[arg-type]
                role="assistant",
                name=self.name,
                messages=new_msgs,
                model_name=model_name,
                message_id=message_id,
                conversation_id=user_msg.conversation_id,
                cost_info=cost_info,
                response_time=time.perf_counter() - start_time,
                provider_response_id=provider_response_id,
                provider_name=provider_name,
                finish_reason=finish_reason,
            )

            # Send additional enriched completion event
            yield StreamCompleteEvent(message=response_msg)
            self.message_sent.emit(response_msg)
            await self.log_message(response_msg)
            if store_history:
                self.conversation.add_chat_messages([user_msg, response_msg])
            await self.connections.route_message(
                response_msg,
                wait=wait_for_connections,
            )

        except Exception as e:
            self.log.exception("Agent stream failed")
            self.run_failed.emit("Agent stream failed", e)
            raise

    async def run_iter(
        self,
        *prompt_groups: Sequence[PromptCompatible],
        output_type: type[OutputDataT] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        wait_for_connections: bool | None = None,
    ) -> AsyncIterator[ChatMessage[OutputDataT]]:
        """Run agent sequentially on multiple prompt groups.

        Args:
            prompt_groups: Groups of prompts to process sequentially
            output_type: Optional type for structured responses
            model: Optional model override
            store_history: Whether to store in conversation history
            wait_for_connections: Whether to wait for connected agents

        Yields:
            Response messages in sequence

        Example:
            questions = [
                ["What is your name?"],
                ["How old are you?", image1],
                ["Describe this image", image2],
            ]
            async for response in agent.run_iter(*questions):
                print(response.content)
        """
        for prompts in prompt_groups:
            response = await self.run(
                *prompts,
                output_type=output_type,
                model=model,
                store_history=store_history,
                wait_for_connections=wait_for_connections,
            )
            yield response  # pyright: ignore

    @method_spawner
    async def run_job(
        self,
        job: Job[TDeps, str | None],
        *,
        store_history: bool = True,
        include_agent_tools: bool = True,
    ) -> ChatMessage[OutputDataT]:
        """Execute a pre-defined task.

        Args:
            job: Job configuration to execute
            store_history: Whether the message exchange should be added to the
                           context window
            include_agent_tools: Whether to include agent tools
        Returns:
            Job execution result

        Raises:
            JobError: If task execution fails
            ValueError: If task configuration is invalid
        """
        if job.required_dependency is not None:  # noqa: SIM102
            if not isinstance(self.context.data, job.required_dependency):
                msg = (
                    f"Agent dependencies ({type(self.context.data)}) "
                    f"don't match job requirement ({job.required_dependency})"
                )
                raise JobError(msg)

        # Load task knowledge
        if job.knowledge:
            # Add knowledge sources to context
            resources: list[Resource | str] = list(job.knowledge.paths) + list(
                job.knowledge.resources
            )
            for source in resources:
                await self.conversation.load_context_source(source)
            for prompt in job.knowledge.prompts:
                await self.conversation.load_context_source(prompt)
        try:
            # Register task tools temporarily
            tools = job.get_tools()
            with self.tools.temporary_tools(tools, exclusive=not include_agent_tools):
                # Execute job with job-specific tools
                return await self.run(await job.get_prompt(), store_history=store_history)

        except Exception as e:
            self.log.exception("Task execution failed", error=str(e))
            msg = f"Task execution failed: {e}"
            raise JobError(msg) from e

    async def run_in_background(
        self,
        *prompt: PromptCompatible,
        max_count: int | None = None,
        interval: float = 1.0,
        block: bool = False,
        **kwargs: Any,
    ) -> ChatMessage[OutputDataT] | None:
        """Run agent continuously in background with prompt or dynamic prompt function.

        Args:
            prompt: Static prompt or function that generates prompts
            max_count: Maximum number of runs (None = infinite)
            interval: Seconds between runs
            block: Whether to block until completion
            **kwargs: Arguments passed to run()
        """
        self._infinite = max_count is None

        async def _continuous():
            count = 0
            self.log.debug(
                "Starting continuous run", max_count=max_count, interval=interval
            )
            latest = None
            while max_count is None or count < max_count:
                try:
                    current_prompts = [
                        call_with_context(p, self.context, **kwargs) if callable(p) else p
                        for p in prompt
                    ]
                    self.log.debug("Generated prompt", iteration=count)
                    latest = await self.run(current_prompts, **kwargs)
                    self.log.debug("Run continuous result", iteration=count)

                    count += 1
                    await asyncio.sleep(interval)
                except asyncio.CancelledError:
                    self.log.debug("Continuous run cancelled")
                    break
                except Exception:
                    self.log.exception("Background run failed")
                    await asyncio.sleep(interval)
            self.log.debug("Continuous run completed", iterations=count)
            return latest

        # Cancel any existing background task
        await self.stop()
        task = asyncio.create_task(_continuous(), name=f"background_{self.name}")
        if block:
            try:
                return await task  # type: ignore
            finally:
                if not task.done():
                    task.cancel()
        else:
            self.log.debug("Started background task", task_name=task.get_name())
            self._background_task = task
            return None

    async def stop(self):
        """Stop continuous execution if running."""
        if self._background_task and not self._background_task.done():
            self._background_task.cancel()
            await self._background_task
            self._background_task = None

    async def wait(self) -> ChatMessage[OutputDataT]:
        """Wait for background execution to complete."""
        if not self._background_task:
            msg = "No background task running"
            raise RuntimeError(msg)
        if self._infinite:
            msg = "Cannot wait on infinite execution"
            raise RuntimeError(msg)
        try:
            return await self._background_task
        finally:
            self._background_task = None

    async def share(
        self,
        target: Agent[TDeps, Any],
        *,
        tools: list[str] | None = None,
        resources: list[str] | None = None,
        history: bool | int | None = None,  # bool or number of messages
        token_limit: int | None = None,
    ):
        """Share capabilities and knowledge with another agent.

        Args:
            target: Agent to share with
            tools: List of tool names to share
            resources: List of resource names to share
            history: Share conversation history:
                    - True: Share full history
                    - int: Number of most recent messages to share
                    - None: Don't share history
            token_limit: Optional max tokens for history

        Raises:
            ValueError: If requested items don't exist
            RuntimeError: If runtime not available for resources
        """
        # Share tools if requested
        for name in tools or []:
            if tool := await self.tools.get_tool(name):
                meta = {"shared_from": self.name}
                target.tools.register_tool(tool.callable, metadata=meta)
            else:
                msg = f"Tool not found: {name}"
                raise ValueError(msg)

        # Share resources if requested
        if resources:
            if not self.runtime:
                msg = "No runtime available for sharing resources"
                raise RuntimeError(msg)
            for name in resources:
                if resource := self.runtime.get_resource(name):
                    await target.conversation.load_context_source(resource)  # type: ignore
                else:
                    msg = f"Resource not found: {name}"
                    raise ValueError(msg)

        # Share history if requested
        if history:
            history_text = await self.conversation.format_history(
                max_tokens=token_limit,
                num_messages=history if isinstance(history, int) else None,
            )
            target.conversation.add_context_message(
                history_text, source=self.name, metadata={"type": "shared_history"}
            )

    def register_worker(
        self,
        worker: MessageNode[Any, Any],
        *,
        name: str | None = None,
        reset_history_on_run: bool = True,
        pass_message_history: bool = False,
    ) -> Tool:
        """Register another agent as a worker tool."""
        return self.tools.register_worker(
            worker,
            name=name,
            reset_history_on_run=reset_history_on_run,
            pass_message_history=pass_message_history,
            parent=self if pass_message_history else None,
        )

    def set_model(self, model: ModelType):
        """Set the model for this agent.

        Args:
            model: New model to use (name or instance)

        """
        self._model = model

    async def reset(self):
        """Reset agent state (conversation history and tool states)."""
        old_tools = await self.tools.list_tools()
        self.conversation.clear()
        self.tools.reset_states()
        new_tools = await self.tools.list_tools()

        event = self.AgentReset(
            agent_name=self.name,
            previous_tools=old_tools,
            new_tools=new_tools,
        )
        self.agent_reset.emit(event)

    @property
    def runtime(self) -> RuntimeConfig:
        """Get runtime configuration from context."""
        assert self.context.runtime
        return self.context.runtime

    @runtime.setter
    def runtime(self, value: RuntimeConfig):
        """Set runtime configuration and update context."""
        self.context.runtime = value

    async def get_stats(self) -> MessageStats:
        """Get message statistics (async version)."""
        messages = await self.get_message_history()
        return MessageStats(messages=messages)

    @asynccontextmanager
    async def temporary_state[T](
        self,
        *,
        system_prompts: list[AnyPromptType] | None = None,
        output_type: type[T] | None = None,
        replace_prompts: bool = False,
        tools: list[ToolType] | None = None,
        replace_tools: bool = False,
        history: list[AnyPromptType] | SessionQuery | None = None,
        replace_history: bool = False,
        pause_routing: bool = False,
        model: ModelType | None = None,
    ) -> AsyncIterator[Self | Agent[T]]:
        """Temporarily modify agent state.

        Args:
            system_prompts: Temporary system prompts to use
            output_type: Temporary output type to use
            replace_prompts: Whether to replace existing prompts
            tools: Temporary tools to make available
            replace_tools: Whether to replace existing tools
            history: Conversation history (prompts or query)
            replace_history: Whether to replace existing history
            pause_routing: Whether to pause message routing
            model: Temporary model override
        """
        old_model = self._model
        if output_type:
            old_type = self._output_type
            self.set_output_type(output_type)  # type: ignore
        async with AsyncExitStack() as stack:
            # System prompts (async)
            if system_prompts is not None:
                await stack.enter_async_context(
                    self.sys_prompts.temporary_prompt(
                        system_prompts, exclusive=replace_prompts
                    )
                )

            # Tools (sync)
            if tools is not None:
                stack.enter_context(
                    self.tools.temporary_tools(tools, exclusive=replace_tools)
                )

            # History (async)
            if history is not None:
                await stack.enter_async_context(
                    self.conversation.temporary_state(
                        history, replace_history=replace_history
                    )
                )

            # Routing (async)
            if pause_routing:
                await stack.enter_async_context(self.connections.paused_routing())

            # Model
            elif model is not None:
                self._model = model

            try:
                yield self
            finally:
                # Restore model
                if model is not None and old_model:
                    self._model = old_model
                if output_type:
                    self.set_output_type(old_type)

    async def validate_against(
        self,
        prompt: str,
        criteria: type[OutputDataT],
        **kwargs: Any,
    ) -> bool:
        """Check if agent's response satisfies stricter criteria."""
        result = await self.run(prompt, **kwargs)
        try:
            criteria.model_validate(result.content.model_dump())  # type: ignore
        except ValidationError:
            return False
        else:
            return True


if __name__ == "__main__":
    import logging

    logfire.configure()
    logfire.instrument_pydantic_ai()
    logging.basicConfig(handlers=[logfire.LogfireLoggingHandler()])
    sys_prompt = "Open browser with google,"
    _model = "openai:gpt-5-nano"

    async def handle_events(ctx, event):
        print(f"[EVENT] {type(event).__name__}: {event}")

    agent = Agent(model=_model, tools=["webbrowser.open"], event_handlers=[handle_events])
    result = agent.run.sync(sys_prompt)



#!/usr/bin/env python3
"""Database migration management script for llmling-agent.

This script provides easy commands for managing database migrations using Alembic.
"""

from __future__ import annotations

import os
from pathlib import Path
import subprocess
from typing import Annotated

from rich.console import Console
from rich.prompt import Confirm
import typer


app = typer.Typer(
    name="llmling-agent-db",
    help="Database migration management for llmling-agent",
    rich_markup_mode="rich",
)
console = Console()


def run_command(cmd: list[str], cwd: Path | None = None) -> int:
    """Run a command and return the exit code."""
    console.print(f"[bold blue]Running:[/bold blue] {' '.join(cmd)}")
    result = subprocess.run(cmd, cwd=cwd, check=False)
    if result.returncode != 0 and cmd[0] == "alembic":
        console.print(
            f"[bold red]Error:[/bold red] Command {cmd[0]!r} not found. "
            "Make sure alembic is installed."
        )
    return result.returncode


def get_project_root() -> Path:
    """Get the project root directory."""
    # Start from this file and go up to find the project root
    # This file is in src/llmling_agent_storage/sql_provider/cli.py
    return Path(__file__).parent.parent.parent.parent


@app.command()
def status() -> int:
    """Show current migration status."""
    project_root = get_project_root()
    return run_command(["alembic", "current", "-v"], cwd=project_root)


@app.command()
def upgrade(
    revision: Annotated[str, typer.Argument(help="Target revision")] = "head",
) -> int:
    """Upgrade database to latest migration."""
    project_root = get_project_root()
    return run_command(["alembic", "upgrade", revision], cwd=project_root)


@app.command()
def downgrade(
    revision: Annotated[
        str,
        typer.Argument(
            help="Target revision (e.g., -1 for previous, or specific revision ID)"
        ),
    ],
) -> int:
    """Downgrade database."""
    project_root = get_project_root()
    return run_command(["alembic", "downgrade", revision], cwd=project_root)


@app.command()
def create(
    message: Annotated[str, typer.Argument(help="Migration message")],
    autogenerate: Annotated[
        bool,
        typer.Option(
            "--autogenerate",
            help="Auto-generate migration based on model changes",
        ),
    ] = False,
) -> int:
    """Create a new migration."""
    project_root = get_project_root()
    cmd = ["alembic", "revision", "-m", message]
    if autogenerate:
        cmd.append("--autogenerate")
    return run_command(cmd, cwd=project_root)


@app.command()
def history() -> int:
    """Show migration history."""
    project_root = get_project_root()
    return run_command(["alembic", "history", "-v"], cwd=project_root)


@app.command()
def current() -> int:
    """Show current revision."""
    project_root = get_project_root()
    return run_command(["alembic", "current"], cwd=project_root)


@app.command()
def reset(
    force: Annotated[
        bool,
        typer.Option("--force", help="Force reset without confirmation"),
    ] = False,
) -> int:
    """Reset database (WARNING: destructive)."""
    if not force and not Confirm.ask(
        "[bold red]This will delete all data in the database. Continue?[/bold red]",
        default=False,
    ):
        console.print("[yellow]Aborted.[/yellow]")
        return 0

    project_root = get_project_root()
    console.print("[bold yellow]Resetting database...[/bold yellow]")

    # First, drop to base (empty database)
    result = run_command(["alembic", "downgrade", "base"], cwd=project_root)
    if result != 0:
        return result

    # Then upgrade to head
    return run_command(["alembic", "upgrade", "head"], cwd=project_root)


def main():
    """Main entry point."""
    # Set up environment
    project_root = get_project_root()
    env = os.environ.copy()
    env["PYTHONPATH"] = str(project_root / "src")
    os.environ.update(env)

    app()


if __name__ == "__main__":
    main()



"""Upsonic based toolset implementation."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from llmling.core.log import get_logger

from llmling_agent.resource_providers.base import ResourceProvider


if TYPE_CHECKING:
    from pydantic import SecretStr

    from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


class UpsonicTools(ResourceProvider):
    """Provider for upsonic tools."""

    def __init__(self, api_key: SecretStr | None = None, base_url: str | None = None):
        from upsonic import Tiger, Tiger_Admin

        super().__init__(name="tiger")
        key = api_key.get_secret_value() if api_key else os.getenv("UPSONIC_API_KEY")
        if base_url:
            assert key
            self.tiger_instance = Tiger_Admin(api_url=base_url, access_key=key)
        else:
            self.tiger_instance = Tiger()
        self._tools: list[Tool] | None = None

    async def get_tools(self) -> list[Tool]:
        """Get tools from entry points."""
        # Return cached tools if available
        if self._tools is not None:
            return self._tools

        self._tools = []
        return self._tools


if __name__ == "__main__":
    import asyncio

    async def main():
        from llmling_agent import Agent

        tools = UpsonicTools()
        agent = Agent(model="gpt-5-nano")
        agent.tools.add_provider(tools)
        result = await agent.run("tell me the tools at your disposal")
        print(result)

    asyncio.run(main())



"""Test ACP filesystem fsspec operations with headless client.

This test validates that the ACP filesystem can be used as a real fsspec
filesystem to perform operations on the client's filesystem using the
headless client with real operations.
"""

from pathlib import Path
import tempfile

import pytest

from acp.filesystem import ACPFileSystem
from llmling_agent_acp.headless_client import HeadlessACPClient


async def test_acp_filesystem_fsspec_operations():
    """Test ACP filesystem with real fsspec operations using headless client."""
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_path = Path(tmpdir)

        # Create test files
        (temp_path / "file1.txt").write_text("Content of file 1")
        (temp_path / "file2.md").write_text("# File 2")

        subdir = temp_path / "subdir"
        subdir.mkdir()
        (subdir / "nested.txt").write_text("Nested content")

        # Create headless client with real operations
        client = HeadlessACPClient(
            working_dir=temp_path,
            allow_file_operations=True,
            auto_grant_permissions=True,
        )

        try:
            session_id = "test_session"
            fs = ACPFileSystem(client=client, session_id=session_id)

            # Test reading files
            content = await fs._cat_file(str(temp_path / "file1.txt"))
            assert content == b"Content of file 1"

            # Test writing files
            new_file = temp_path / "written.txt"
            await fs._put_file(str(new_file), "New content")
            assert new_file.exists()
            assert new_file.read_text() == "New content"

            # Test directory listing
            files = await fs._ls(str(temp_path), detail=True)
            file_names = [f["name"] for f in files]

            assert "file1.txt" in file_names
            assert "file2.md" in file_names
            assert "subdir" in file_names
            assert "written.txt" in file_names

            # Test file existence
            assert await fs._exists(str(temp_path / "file1.txt")) is True
            assert await fs._exists(str(temp_path / "nonexistent.txt")) is False

            # Test isfile/isdir
            assert await fs._isfile(str(temp_path / "file1.txt")) is True
            assert await fs._isdir(str(temp_path / "subdir")) is True
            assert await fs._isfile(str(temp_path / "subdir")) is False

            # Test file info
            info = await fs._info(str(temp_path / "file1.txt"))
            assert info["name"] == "file1.txt"
            assert info["type"] == "file"
            assert info["size"] == 17  # Length of "Content of file 1"  # noqa: PLR2004

        finally:
            await client.cleanup()


#
# async def test_acp_filesystem_with_nested_directories():
#     """Test ACP filesystem operations with nested directory structures."""
#     with tempfile.TemporaryDirectory() as tmpdir:
#         temp_path = Path(tmpdir)

#         # Create nested structure
#         (temp_path / "root.txt").write_text("Root file")

#         level1 = temp_path / "level1"
#         level1.mkdir()
#         (level1 / "l1_file.txt").write_text("Level 1 file")

#         level2 = level1 / "level2"
#         level2.mkdir()
#         (level2 / "l2_file.txt").write_text("Level 2 file")

#         client = HeadlessACPClient(
#             working_dir=temp_path,
#             allow_file_operations=True,
#             auto_grant_permissions=True,
#         )

#         try:
#             fs = ACPFileSystem(client=client, session_id="test_session")

#             # Test reading from nested directories
#             content = await fs._cat_file(str(level2 / "l2_file.txt"))
#             assert content == b"Level 2 file"

#             # Test listing nested directories
#             level1_files = await fs._ls(str(level1), detail=True)
#             level1_names = [f["name"] for f in level1_files]
#             assert "l1_file.txt" in level1_names
#             assert "level2" in level1_names

#             # Test creating files in nested dirs
#             new_nested_file = level2 / "created.txt"
#             await fs._put_file(str(new_nested_file), "Created in nested dir")
#             assert new_nested_file.exists()
#             assert new_nested_file.read_text() == "Created in nested dir"

#             print("✅ Nested directory operations working!")

#         finally:
#             await client.cleanup()


#
# async def test_acp_filesystem_error_conditions():
#     """Test ACP filesystem error handling."""
#     with tempfile.TemporaryDirectory() as tmpdir:
#         temp_path = Path(tmpdir)

#         client = HeadlessACPClient(
#             working_dir=temp_path,
#             allow_file_operations=True,
#             auto_grant_permissions=True,
#         )

#         try:
#             fs = ACPFileSystem(client=client, session_id="test_session")

#             # Test reading non-existent file
#             with pytest.raises(FileNotFoundError):
#                 await fs._cat_file(str(temp_path / "nonexistent.txt"))

#             # Test file existence for non-existent file
#             assert await fs._exists(str(temp_path / "nonexistent.txt")) is False

#             # Test listing non-existent directory
#             with pytest.raises(FileNotFoundError):
#                 await fs._ls(str(temp_path / "nonexistent_dir"))

#             print("✅ Error handling working correctly!")

#         finally:
#             await client.cleanup()


if __name__ == "__main__":
    pytest.main(["-v", "-s", __file__])



"""FastMCP-based client implementation for LLMling agent.

This module provides a client for communicating with MCP servers using FastMCP.
It includes support for contextual progress handlers that extend FastMCP's
standard progress callbacks with tool execution context (tool name, call ID, and input).

The key innovation is the signature injection system that allows MCP tools to work
seamlessly with PydanticAI's RunContext while providing rich progress information.
"""

from __future__ import annotations

import base64
from collections.abc import Callable
import contextlib
import inspect
import logging
from typing import TYPE_CHECKING, Any, Self

from anyenv import MultiEventHandler
from pydantic_ai import (
    BinaryContent,
    BinaryImage,
    DocumentUrl,
    RunContext,  # noqa: TC002
    ToolReturn,
)
from schemez.functionschema import FunctionSchema

from llmling_agent.log import get_logger
from llmling_agent.mcp_server.helpers import (
    _create_tool_annotations_with_context,
    _create_tool_signature_with_context,
    extract_text_content,
    extract_tool_call_args,
    mcp_tool_to_fn_schema,
)
from llmling_agent.mcp_server.message_handler import MCPMessageHandler
from llmling_agent.tools.base import Tool
from llmling_agent_config.mcp_server import (
    SSEMCPServerConfig,
    StdioMCPServerConfig,
    StreamableHTTPMCPServerConfig,
)


type ContextualProgressHandler = Callable[
    [
        float,  # progress
        float | None,  # total
        str | None,  # message
        str | None,  # tool_name
        str | None,  # tool_call_id
        dict[str, Any] | None,  # tool_input
    ],
    Any,  # Supports both sync (None) and async (Awaitable[None]) returns
]


if TYPE_CHECKING:
    from collections.abc import Sequence
    from types import TracebackType

    from fastmcp.client import ClientTransport
    from fastmcp.client.client import ProgressHandler
    from fastmcp.client.elicitation import ElicitationHandler, ElicitResult
    from fastmcp.client.logging import LogMessage
    from fastmcp.client.messages import MessageHandler, MessageHandlerT
    from fastmcp.client.sampling import ClientSamplingHandler
    import mcp
    from mcp.client.session import RequestContext
    from mcp.types import (
        BlobResourceContents,
        ContentBlock,
        CreateMessageRequestParams,
        ElicitRequestParams,
        Prompt as MCPPrompt,
        Resource as MCPResource,
        SamplingMessage,
        TextResourceContents,
        Tool as MCPTool,
    )

    from llmling_agent_config.mcp_server import MCPServerConfig


logger = get_logger(__name__)


LEVEL_MAP = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "notice": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "alert": logging.CRITICAL,
    "emergency": logging.CRITICAL,
}


class MCPClient:
    """FastMCP-based client for communicating with MCP servers."""

    def __init__(
        self,
        config: MCPServerConfig,
        elicitation_callback: ElicitationHandler | None = None,
        sampling_callback: ClientSamplingHandler | None = None,
        progress_handler: ContextualProgressHandler | None = None,
        message_handler: MessageHandlerT | MessageHandler | None = None,
        accessible_roots: list[str] | None = None,
    ):
        self._elicitation_callback = elicitation_callback
        self.config = config
        self._sampling_callback = sampling_callback
        self._progress_handler = MultiEventHandler[ContextualProgressHandler](
            handlers=[progress_handler] if progress_handler else []
        )
        # Store message handler or mark for lazy creation
        self._message_handler = message_handler
        self._accessible_roots = accessible_roots or []
        self._client = self._get_client(self.config)
        self._available_tools: list[MCPTool] = []
        self._connected = False

    async def __aenter__(self) -> Self:
        """Enter context manager."""
        try:
            # First attempt with configured auth
            await self._client.__aenter__()

        except Exception as first_error:
            # OAuth fallback for HTTP/SSE if not already using OAuth
            if _should_try_oauth_fallback(self.config):
                try:
                    with contextlib.suppress(Exception):
                        await self._client.__aexit__(None, None, None)
                    self._client = self._get_client(self.config, force_oauth=True)
                    await self._client.__aenter__()
                    logger.info("Connected with OAuth fallback")
                except Exception:  # noqa: BLE001
                    raise first_error from None
            else:
                raise

        self._connected = True
        await self._refresh_tools()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        """Exit context manager and cleanup."""
        try:
            await self._client.__aexit__(None, None, None)
        except Exception as e:  # noqa: BLE001
            logger.warning("Error during FastMCP client cleanup", error=e)
        finally:
            self._connected = False
            self._available_tools = []

    async def _log_handler(self, message: LogMessage) -> None:
        """Handle server log messages."""
        level = LEVEL_MAP.get(message.level.lower(), logging.INFO)
        logger.log(level, "MCP Server: ", data=message.data)

    async def _elicitation_handler_impl(
        self,
        message: str,
        response_type: type,
        params: ElicitRequestParams,
        context: RequestContext,
    ) -> ElicitResult[dict[str, Any]] | dict[str, Any] | None:
        """Handle elicitation requests from server."""
        if not self._elicitation_callback:
            return None

        try:
            # Direct FastMCP callback - no compatibility layer
            return await self._elicitation_callback(
                message, response_type, params, context
            )
        except Exception:
            logger.exception("Elicitation handler failed")
            from fastmcp.client.elicitation import ElicitResult

            return ElicitResult(action="cancel")

    async def _sampling_handler_impl(
        self,
        messages: list[SamplingMessage],
        params: CreateMessageRequestParams,
        context: RequestContext,
    ) -> str:
        """Handle sampling requests from server."""
        if not self._sampling_callback:
            return "Sampling not supported"

        try:
            result = self._sampling_callback(messages, params, context)
            if inspect.iscoroutine(result):
                result = await result
            return str(result)
        except Exception as e:
            logger.exception("Sampling handler failed")
            return f"Sampling failed: {e}"

    def _get_client(self, config: MCPServerConfig, force_oauth: bool = False):
        """Create FastMCP client based on config."""
        import fastmcp
        from fastmcp.client import SSETransport, StreamableHttpTransport
        from fastmcp.client.transports import StdioTransport

        transport: ClientTransport
        # Create transport based on config type
        match config:
            case StdioMCPServerConfig(command=command, args=args):
                env = config.get_env_vars()
                transport = StdioTransport(command=command, args=args, env=env)
                oauth = False
                if force_oauth:
                    msg = "OAuth is not supported for StdioMCPServerConfig"
                    raise ValueError(msg)

            case SSEMCPServerConfig(url=url, headers=headers, auth=auth):
                transport = SSETransport(url=url, headers=headers)
                oauth = auth.oauth

            case StreamableHTTPMCPServerConfig(url=url, headers=headers, auth=auth):
                transport = StreamableHttpTransport(url=url, headers=headers)
                oauth = auth.oauth
            case _:
                msg = f"Unsupported server config type: {type(config)}"
                raise ValueError(msg)

        # Create message handler if needed
        msg_handler = self._message_handler or MCPMessageHandler(self)
        return fastmcp.Client(
            transport,
            log_handler=self._log_handler,
            roots=self._accessible_roots,
            timeout=config.timeout,
            elicitation_handler=self._elicitation_handler_impl,
            sampling_handler=self._sampling_handler_impl,
            message_handler=msg_handler,
            auth="oauth" if (force_oauth or oauth) else None,
        )

    async def _refresh_tools(self) -> None:
        """Refresh the list of available tools from the server."""
        if not self._client or not self._connected:
            return

        try:
            tools = await self._client.list_tools()
            self._available_tools = tools
            logger.debug("Refreshed tools from MCP server", num_tools=len(tools))
        except Exception as e:  # noqa: BLE001
            logger.warning("Failed to refresh tools", error=e)
            self._available_tools = []

    def get_tools(self) -> list[dict[str, Any]]:
        """Get tools in OpenAI function format."""
        return [
            {"type": "function", "function": mcp_tool_to_fn_schema(tool)}
            for tool in self._available_tools
        ]

    async def list_prompts(self) -> list[MCPPrompt]:
        """Get available prompts from the server."""
        if not self._client or not self._connected:
            msg = "Not connected to MCP server"
            raise RuntimeError(msg)

        try:
            return await self._client.list_prompts()
        except Exception as e:  # noqa: BLE001
            logger.debug("Failed to list prompts", error=e)
            return []

    async def list_resources(self) -> list[MCPResource]:
        """Get available resources from the server."""
        if not self._client or not self._connected:
            msg = "Not connected to MCP server"
            raise RuntimeError(msg)

        try:
            return await self._client.list_resources()
        except Exception as e:
            msg = f"Failed to list resources: {e}"
            raise RuntimeError(msg) from e

    async def get_prompt(
        self, name: str, arguments: dict[str, str] | None = None
    ) -> mcp.types.GetPromptResult:
        """Get a specific prompt's content."""
        if not self._client or not self._connected:
            msg = "Not connected to MCP server"
            raise RuntimeError(msg)

        try:
            return await self._client.get_prompt_mcp(name, arguments)
        except Exception as e:
            msg = f"Failed to get prompt {name!r}: {e}"
            raise RuntimeError(msg) from e

    def convert_tool(self, tool: MCPTool) -> Tool:
        """Create a properly typed callable from MCP tool schema."""
        schema = mcp_tool_to_fn_schema(tool)
        fn_schema = FunctionSchema.from_dict(schema)
        sig = fn_schema.to_python_signature()

        async def tool_callable(ctx: RunContext, **kwargs: Any) -> str | Any | ToolReturn:
            """Dynamically generated MCP tool wrapper."""
            # Filter out None values for optional params
            schema_props = tool.inputSchema.get("properties", {})
            required_props = set(tool.inputSchema.get("required", []))
            filtered_kwargs = {
                k: v
                for k, v in kwargs.items()
                if k in required_props or (k in schema_props and v is not None)
            }
            return await self.call_tool(
                tool.name,
                filtered_kwargs,
                tool_call_id=ctx.tool_call_id if ctx else None,
                run_context=ctx,
            )

        # Set proper signature and annotations with RunContext support
        tool_callable.__signature__ = _create_tool_signature_with_context(sig)  # type: ignore
        annotations = _create_tool_annotations_with_context(fn_schema.get_annotations())
        # Update return annotation to support multiple types
        annotations["return"] = str | Any | ToolReturn
        tool_callable.__annotations__ = annotations
        tool_callable.__name__ = tool.name
        tool_callable.__doc__ = tool.description or "No description provided."
        return Tool.from_callable(tool_callable, source="mcp")

    def _create_final_progress_handler(
        self, tool_name: str, tool_call_id: str, tool_input: dict[str, Any]
    ) -> ProgressHandler:
        """Create a FastMCP-compatible progress handler with baked-in context."""

        async def fastmcp_progress_handler(
            progress: float, total: float | None, message: str | None
        ):
            await self._progress_handler(
                progress, total, message, tool_name, tool_call_id, tool_input
            )

        return fastmcp_progress_handler

    async def call_tool(
        self,
        name: str,
        arguments: dict | None = None,
        tool_call_id: str | None = None,
        run_context: RunContext | None = None,
    ) -> ToolReturn | str | Any:
        """Call an MCP tool with full PydanticAI return type support."""
        if not self._client or not self._connected:
            msg = "Not connected to MCP server"
            raise RuntimeError(msg)

        # Create progress handler if we have handler
        progress_handler = None
        if self._progress_handler:
            if run_context and run_context.tool_call_id and run_context.tool_name:
                # Extract tool args from message history
                tool_input = extract_tool_call_args(
                    run_context.messages, run_context.tool_call_id
                )
                progress_handler = self._create_final_progress_handler(
                    run_context.tool_name, run_context.tool_call_id, tool_input
                )
            else:
                # Fallback to using passed arguments (direct tool call)
                progress_handler = self._create_final_progress_handler(
                    name, tool_call_id or "", arguments or {}
                )
        try:
            result = await self._client.call_tool(
                name, arguments, progress_handler=progress_handler
            )
            content = await self._convert_mcp_content(result.content)
            # Decision logic for return type
            match (result.data is not None, bool(content)):
                case (True, True):  # Both structured data and rich content -> ToolReturn
                    return ToolReturn(return_value=result.data, content=content)
                case (True, False):  # Only structured data -> return directly
                    return result.data
                case (False, True):  # Only content -> ToolReturn with content
                    msg = "Tool executed successfully"
                    return ToolReturn(return_value=msg, content=content)
                case (False, False):  # Fallback to text extraction
                    return extract_text_content(result.content)
                case _:  # Handle unexpected cases
                    msg = f"Unexpected MCP content: {result.content}"
                    raise ValueError(msg)  # noqa: TRY301
        except Exception as e:
            msg = f"MCP tool call failed: {e}"
            raise RuntimeError(msg) from e

    async def _convert_mcp_content(
        self,
        mcp_content: Sequence[ContentBlock | TextResourceContents | BlobResourceContents],
    ) -> list[str | BinaryContent]:
        """Convert MCP content blocks to PydanticAI content types."""
        from mcp.types import (
            AudioContent,
            BlobResourceContents,
            EmbeddedResource,
            ImageContent,
            ResourceLink,
            TextContent,
            TextResourceContents,
        )

        contents: list[Any] = []

        for block in mcp_content:
            match block:
                case TextContent(text=text):
                    contents.append(text)
                case TextResourceContents(text=text):
                    contents.append(text)
                case ImageContent(data=data, mimeType=mime_type):
                    decoded_data = base64.b64decode(data)
                    img = BinaryImage(data=decoded_data, media_type=mime_type)
                    contents.append(img)
                case AudioContent(data=data, mimeType=mime_type):
                    decoded_data = base64.b64decode(data)
                    content = BinaryContent(data=decoded_data, media_type=mime_type)
                    contents.append(content)
                case BlobResourceContents(blob=blob):
                    decoded_data = base64.b64decode(blob)
                    mime = "application/octet-stream"
                    content = BinaryContent(data=decoded_data, media_type=mime)
                    contents.append(content)
                case ResourceLink(uri=uri):
                    try:
                        assert self._client
                        res = await self._client.read_resource_mcp(uri)
                        nested = await self._convert_mcp_content(res.contents)
                        contents.extend(nested)
                    except Exception:  # noqa: BLE001
                        # Fallback to DocumentUrl if reading fails
                        contents.append(DocumentUrl(url=str(uri)))
                case EmbeddedResource(resource=TextResourceContents(text=text)):
                    contents.append(text)
                case EmbeddedResource(resource=BlobResourceContents() as blob_resource):
                    contents.append(f"[Binary data: {blob_resource.mimeType}]")
                case _:
                    contents.append(str(block))  # Convert anything else to string

        return contents


def _should_try_oauth_fallback(config: MCPServerConfig) -> bool:
    """Check if OAuth fallback should be attempted."""
    # No fallback for stdio or if OAuth already configured
    return not isinstance(config, StdioMCPServerConfig) and not config.auth.oauth



"""Simplified observability registry using Logfire with single backend."""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import logfire

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from llmling_agent_config.observability import (
        BaseObservabilityConfig,
        ObservabilityConfig,
    )

logger = get_logger(__name__)


class ObservabilityRegistry:
    """Simplified registry that configures Logfire for single backend export."""

    def __init__(self):
        self._configured = False

    def configure_observability(self, observability_config: ObservabilityConfig):
        """Configure Logfire for single backend export.

        Args:
            observability_config: Configuration for observability
        """
        if not observability_config.enabled or not observability_config.provider:
            logger.debug("Observability disabled or no provider configured")
            return

        if self._configured:
            logger.warning("Observability already configured, skipping")
            return

        config = observability_config.provider
        if not config.enabled:
            logger.debug("Provider is disabled", provider=config.type)
            return

        _setup_otel_environment(config)  # Configure OTEL env variables based on provider
        logfire.configure(
            service_name=config.service_name,
            environment=config.environment,
            console=False,
            send_to_logfire=(config.type == "logfire"),
        )
        logfire.instrument_pydantic_ai()
        logging.basicConfig(handlers=[logfire.LogfireLoggingHandler()])

        self._configured = True
        logger.info("Configured observability", provider=config.type)


def _setup_otel_environment(config: BaseObservabilityConfig):
    """Set up OTEL environment variables for the configured backend."""
    # Get endpoint and headers from config
    endpoint = getattr(config, "_endpoint", getattr(config, "endpoint", None))
    headers = getattr(config, "_headers", getattr(config, "headers", {}))

    if not endpoint:
        logger.warning("No endpoint found", provider=config.type)
        return

    # Set standard OTEL environment variables
    os.environ["OTEL_EXPORTER_OTLP_ENDPOINT"] = endpoint
    os.environ["OTEL_EXPORTER_OTLP_PROTOCOL"] = config.protocol

    # Set headers if available
    if headers:
        header_str = ",".join(f"{k}={v}" for k, v in headers.items())
        os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = header_str

    # Set resource attributes
    resource_attrs = []
    if config.service_name:
        resource_attrs.append(f"service.name={config.service_name}")
    if config.environment:
        resource_attrs.append(f"deployment.environment.name={config.environment}")

    if resource_attrs:
        os.environ["OTEL_RESOURCE_ATTRIBUTES"] = ",".join(resource_attrs)

    logger.debug(
        "Set OTEL environment",
        endpoint=endpoint,
        protocol=config.protocol,
        headers=headers,
    )


# Global registry instance
registry = ObservabilityRegistry()



"""Agent session slash commands."""

from __future__ import annotations

from slashed import CommandContext, SlashedCommand  # noqa: TC002

from llmling_agent.agent.context import AgentContext  # noqa: TC001


class ClearCommand(SlashedCommand):
    """Clear the current chat session history.

    This removes all previous messages but keeps tools and settings.
    """

    name = "clear"
    category = "session"

    async def execute_command(self, ctx: CommandContext[AgentContext]):
        """Clear chat history.

        Args:
            ctx: Command context
        """
        ctx.context.agent.conversation.clear()
        await ctx.print("🧹 **Chat history cleared**")


class ResetCommand(SlashedCommand):
    """Reset the entire session state.

    - Clears chat history
    - Restores default tool settings
    - Resets any session-specific configurations
    """

    name = "reset"
    category = "session"

    async def execute_command(self, ctx: CommandContext[AgentContext]):
        """Reset session state.

        Args:
            ctx: Command context
        """
        await ctx.context.agent.reset()
        await ctx.print(
            "🔄 **Session state reset** - history cleared, tools and settings restored"
        )



"""Tool to chain multiple function calls."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from enum import StrEnum
from typing import Any, Literal

from pydantic import Field
from pydantic_ai import ModelRetry
from schemez import Schema

from llmling_agent.agent.context import AgentContext  # noqa: TC001


class ErrorStrategy(StrEnum):
    """Strategy for handling errors in the pipeline."""

    STOP = "stop"  # Stop pipeline on error
    SKIP = "skip"  # Skip failed step, continue with previous result
    DEFAULT = "default"  # Use provided default value
    RETRY = "retry"  # Retry the step


class StepCondition(Schema):
    """Condition for conditional execution."""

    field: str  # Field to check in result
    operator: Literal["eq", "gt", "lt", "contains", "exists"]
    value: Any = None

    def evaluate_with_value(self, value: Any) -> bool:
        """Evaluate this condition against a value.

        Args:
            value: The value to evaluate against the condition.

        Returns:
            bool: True if the condition is met, False otherwise.
        """
        field_value = value.get(self.field) if isinstance(value, dict) else value

        match self.operator:
            case "eq":
                return field_value == self.value
            case "gt":
                return field_value > self.value
            case "lt":
                return field_value < self.value
            case "contains":
                try:
                    return self.value in field_value  # type: ignore
                except TypeError:
                    return False
            case "exists":
                return field_value is not None
            case _:
                msg = f"Unknown operator: {self.operator}"
                raise ValueError(msg)


@dataclass
class StepResult:
    """Result of a pipeline step execution."""

    success: bool
    result: Any
    error: Exception | None = None
    retries: int = 0
    duration: float = 0.0


# Type alias for step results during execution
type StepResults = dict[str, StepResult]


class PipelineStep(Schema):
    """Single step in a tool pipeline."""

    tool: str
    input_kwarg: str = "text"
    keyword_args: dict[str, Any] = Field(default_factory=dict)
    name: str | None = None  # Optional step name for referencing
    condition: StepCondition | None = None  # Conditional execution
    error_strategy: ErrorStrategy = ErrorStrategy.STOP
    default_value: Any = None  # Used with ErrorStrategy.DEFAULT
    max_retries: int = 0
    retry_delay: float = 1.0
    timeout: float | None = None
    depends_on: list[str] = Field(default_factory=list)  # Step dependencies


class Pipeline(Schema):
    """A pipeline of tool operations."""

    input: str | dict[str, Any]
    steps: list[PipelineStep]
    mode: Literal["sequential", "parallel"] = "sequential"
    max_parallel: int = 5  # Max concurrent steps
    collect_metrics: bool = False  # Collect execution metrics


async def _execute_step(
    ctx: AgentContext,
    step: PipelineStep,
    input_value: Any,
    results: StepResults,
) -> StepResult:
    """Execute a single pipeline step."""
    start_time = asyncio.get_event_loop().time()
    retries = 0

    while True:
        try:
            # Check condition if any
            if step.condition and not step.condition.evaluate_with_value(input_value):
                return StepResult(success=True, result=input_value, duration=0)

            # Get the tool
            tool_info = await ctx.agent.tools.get_tool(step.tool)
            if not tool_info:
                msg = f"Tool {step.tool} not found"
                raise ValueError(msg)  # noqa: TRY301

            # Prepare kwargs
            if isinstance(input_value, dict):
                kwargs = {**input_value, **step.keyword_args}
            else:
                kwargs = {step.input_kwarg: input_value, **step.keyword_args}

            # Execute with timeout if specified
            if step.timeout:
                fut = tool_info.execute(ctx, **kwargs)
                result = await asyncio.wait_for(fut, timeout=step.timeout)
            else:
                result = await tool_info.execute(ctx, **kwargs)

            duration = asyncio.get_event_loop().time() - start_time
            return StepResult(success=True, result=result, duration=duration)

        except Exception as exc:
            match step.error_strategy:
                case ErrorStrategy.STOP:
                    raise

                case ErrorStrategy.SKIP:
                    duration = asyncio.get_event_loop().time() - start_time
                    return StepResult(
                        success=False,
                        result=input_value,
                        error=exc,
                        duration=duration,
                    )

                case ErrorStrategy.DEFAULT:
                    duration = asyncio.get_event_loop().time() - start_time
                    return StepResult(
                        success=False,
                        result=step.default_value,
                        error=exc,
                        duration=duration,
                    )

                case ErrorStrategy.RETRY:
                    retries += 1
                    if retries <= step.max_retries:
                        await asyncio.sleep(step.retry_delay)
                        continue
                    raise  # Max retries exceeded


async def _execute_sequential(
    ctx: AgentContext, pipeline: Pipeline, results: StepResults
) -> Any:
    """Execute steps sequentially."""
    current = pipeline.input

    for step in pipeline.steps:
        result = await _execute_step(ctx, step, current, results)
        if step.name:
            results[step.name] = result
        current = result.result

    return current


async def _execute_parallel(
    ctx: AgentContext, pipeline: Pipeline, results: StepResults
) -> Any:
    """Execute independent steps in parallel."""
    semaphore = asyncio.Semaphore(pipeline.max_parallel)

    async def run_step(step: PipelineStep):
        async with semaphore:
            # Wait for dependencies
            for dep in step.depends_on:
                while dep not in results:
                    await asyncio.sleep(0.1)

            # Get input from dependency or pipeline input
            if step.depends_on:
                input_value = results[step.depends_on[-1]].result
            else:
                input_value = pipeline.input

            result = await _execute_step(ctx, step, input_value, results)
            if step.name:
                results[step.name] = result

    # Create tasks for all steps
    tasks = [run_step(step) for step in pipeline.steps]
    await asyncio.gather(*tasks)

    # Return last result
    return results[name].result if (name := pipeline.steps[-1].name) else None


async def chain_tools(
    ctx: AgentContext,
    input_data: str | dict[str, Any],
    steps: list[dict[str, Any]],
    mode: Literal["sequential", "parallel"] = "sequential",
    max_parallel: int = 5,
    collect_metrics: bool = False,
) -> Any:
    """Execute multiple tool operations in sequence or parallel.

    WHEN TO USE THIS TOOL:
    - Use this when you can plan multiple operations confidently in advance
    - Use this for common sequences you've successfully used before
    - Use this to reduce interaction rounds for known operation patterns
    - Use this when all steps are independent of intermediate results

    DO NOT USE THIS TOOL:
    - When you need to inspect intermediate results
    - When next steps depend on analyzing previous results
    - When you're unsure about the complete sequence
    - When you need to handle errors at each step individually

    Args:
        ctx: Agent context for tool execution
        input_data: Initial input for the pipeline
        steps: List of step configurations, each containing:
            - tool: Name of the tool to execute
            - input_kwarg: Keyword argument name for input (default: "text")
            - keyword_args: Additional keyword arguments
            - name: Optional step name for referencing
            - condition: Optional execution condition
            - error_strategy: How to handle errors ("stop", "skip", "default", "retry")
            - default_value: Value to use with "default" error strategy
            - max_retries: Maximum retry attempts
            - retry_delay: Delay between retries in seconds
            - timeout: Step timeout in seconds
            - depends_on: List of step names this depends on
        mode: Execution mode - "sequential" or "parallel"
        max_parallel: Maximum concurrent steps for parallel mode
        collect_metrics: Whether to collect execution metrics

    Examples:
        # Sequential processing
        await chain_tools(
            ctx,
            input_data="main.py",
            steps=[
                {"tool": "load_resource", "input_kwarg": "name"},
                {"tool": "analyze_code", "input_kwarg": "code"},
                {"tool": "format_output", "input_kwarg": "text"}
            ]
        )

        # Parallel processing with dependencies
        await chain_tools(
            ctx,
            input_data="test.py",
            mode="parallel",
            steps=[
                {"tool": "load_resource", "input_kwarg": "name", "name": "load"},
                {"tool": "analyze_code", "input_kwarg": "code", "depends_on": ["load"]},
                {"tool": "count_tokens", "input_kwarg": "text", "depends_on": ["load"]}
            ]
        )
    """
    try:
        pipeline = Pipeline(
            input=input_data,
            steps=[PipelineStep.model_validate(step) for step in steps],
            mode=mode,
            max_parallel=max_parallel,
            collect_metrics=collect_metrics,
        )
    except Exception as e:
        msg = f"Invalid pipeline configuration: {e}"
        raise ModelRetry(msg) from e
    results: StepResults = {}

    try:
        match pipeline.mode:
            case "sequential":
                return await _execute_sequential(ctx, pipeline, results)
            case "parallel":
                return await _execute_parallel(ctx, pipeline, results)
    except Exception as e:
        msg = f"Failed to execute pipeline: {e}"
        raise ModelRetry(msg) from e



"""Agent ACP Connection."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Self

from acp.agent.protocol import Agent
from acp.client.protocol import Client
from acp.connection import Connection
from acp.exceptions import RequestError
from acp.schema import (
    AuthenticateRequest,
    CancelNotification,
    CreateTerminalRequest,
    CreateTerminalResponse,
    InitializeRequest,
    KillTerminalCommandRequest,
    KillTerminalCommandResponse,
    LoadSessionRequest,
    NewSessionRequest,
    PromptRequest,
    ReadTextFileRequest,
    ReadTextFileResponse,
    ReleaseTerminalRequest,
    ReleaseTerminalResponse,
    RequestPermissionRequest,
    RequestPermissionResponse,
    SessionNotification,
    SetSessionModelRequest,
    SetSessionModeRequest,
    TerminalOutputRequest,
    TerminalOutputResponse,
    WaitForTerminalExitRequest,
    WaitForTerminalExitResponse,
    WriteTextFileRequest,
    WriteTextFileResponse,
)
from acp.task import DebuggingMessageStateStore


if TYPE_CHECKING:
    import asyncio
    from collections.abc import Callable

    from acp.agent.protocol import Agent
    from acp.connection import StreamObserver
    from acp.schema import (
        AgentMethod,
        CreateTerminalRequest,
        InitializeResponse,
        KillTerminalCommandRequest,
        NewSessionResponse,
        PromptResponse,
        ReadTextFileRequest,
        ReleaseTerminalRequest,
        RequestPermissionRequest,
        SessionNotification,
        TerminalOutputRequest,
        WaitForTerminalExitRequest,
        WriteTextFileRequest,
    )


class AgentSideConnection(Client):
    """Agent-side connection.

    Use when you implement the Agent and need to talk to a Client.

    Args:
        to_agent: factory that receives this connection and returns your Agent
        input: asyncio.StreamWriter (local -> peer)
        output: asyncio.StreamReader (peer -> local)
    """

    def __init__(
        self,
        to_agent: Callable[[AgentSideConnection], Agent],
        input_stream: asyncio.StreamWriter,
        output_stream: asyncio.StreamReader,
        observers: list[StreamObserver] | None = None,
        *,
        debug_file: str | None = None,
    ) -> None:
        agent = to_agent(self)
        handler = partial(_agent_handler, agent)
        store = DebuggingMessageStateStore(debug_file=debug_file) if debug_file else None
        self._conn = Connection(
            handler,
            input_stream,
            output_stream,
            state_store=store,
            observers=observers,
        )

    # client-bound methods (agent -> client)
    async def session_update(self, params: SessionNotification[Any]) -> None:
        dct = params.model_dump(by_alias=True, exclude_none=True)
        await self._conn.send_notification("session/update", dct)

    async def request_permission(
        self, params: RequestPermissionRequest
    ) -> RequestPermissionResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        method = "session/request_permission"
        resp = await self._conn.send_request(method, dct)
        return RequestPermissionResponse.model_validate(resp)

    async def read_text_file(self, params: ReadTextFileRequest) -> ReadTextFileResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("fs/read_text_file", dct)
        return ReadTextFileResponse.model_validate(resp)

    async def write_text_file(
        self, params: WriteTextFileRequest
    ) -> WriteTextFileResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        r = await self._conn.send_request("fs/write_text_file", dct)
        return WriteTextFileResponse.model_validate(r)

    # async def createTerminal(self, params: CreateTerminalRequest) -> TerminalHandle:
    async def create_terminal(
        self, params: CreateTerminalRequest
    ) -> CreateTerminalResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("terminal/create", dct)
        #  resp = CreateTerminalResponse.model_validate(resp)
        #  return TerminalHandle(resp.terminal_id, params.session_id, self._conn)
        return CreateTerminalResponse.model_validate(resp)

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        return await self._conn.send_request(f"_{method}", params)

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        await self._conn.send_notification(f"_{method}", params)

    async def terminal_output(
        self, params: TerminalOutputRequest
    ) -> TerminalOutputResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("terminal/output", dct)
        return TerminalOutputResponse.model_validate(resp)

    async def release_terminal(
        self, params: ReleaseTerminalRequest
    ) -> ReleaseTerminalResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("terminal/release", dct)
        return ReleaseTerminalResponse.model_validate(resp)

    async def wait_for_terminal_exit(
        self, params: WaitForTerminalExitRequest
    ) -> WaitForTerminalExitResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("terminal/wait_for_exit", dct)
        return WaitForTerminalExitResponse.model_validate(resp)

    async def kill_terminal(
        self, params: KillTerminalCommandRequest
    ) -> KillTerminalCommandResponse:
        dct = params.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
        resp = await self._conn.send_request("terminal/kill", dct)
        return KillTerminalCommandResponse.model_validate(resp)

    async def close(self) -> None:
        await self._conn.close()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()


async def _agent_handler(  # noqa: PLR0911
    agent: Agent,
    method: AgentMethod | str,
    params: dict[str, Any] | None,
    is_notification: bool,
) -> NewSessionResponse | InitializeResponse | PromptResponse | dict[str, Any] | None:
    match method:
        case "initialize":
            initialize_request = InitializeRequest.model_validate(params)
            return await agent.initialize(initialize_request)
        case "session/new":
            new_session_request = NewSessionRequest.model_validate(params)
            return await agent.new_session(new_session_request)
        case "session/load":
            load_request = LoadSessionRequest.model_validate(params)
            await agent.load_session(load_request)
            return None
        case "session/set_mode":
            set_mode_request = SetSessionModeRequest.model_validate(params)
            return (
                session_resp.model_dump(by_alias=True, exclude_none=True)
                if (session_resp := await agent.set_session_mode(set_mode_request))
                else {}
            )
        case "session/prompt":
            prompt_request = PromptRequest.model_validate(params)
            return await agent.prompt(prompt_request)
        case "session/cancel":
            cancel_notification = CancelNotification.model_validate(params)
            await agent.cancel(cancel_notification)
            return None
        case "session/set_model":
            set_model_request = SetSessionModelRequest.model_validate(params)
            return (
                model_result.model_dump(by_alias=True, exclude_none=True)
                if (model_result := await agent.set_session_model(set_model_request))
                else {}
            )
        case "authenticate":
            p = AuthenticateRequest.model_validate(params)
            result = await agent.authenticate(p)
            return result.model_dump(by_alias=True, exclude_none=True) if result else {}
        case str() if method.startswith("_") and is_notification:
            await agent.ext_notification(method[1:], params or {})
            return None
        case str() if method.startswith("_"):
            return await agent.ext_method(method[1:], params or {})
        case _:
            raise RequestError.method_not_found(method)



"""Base class for converters."""

from __future__ import annotations

from abc import ABC, abstractmethod
import mimetypes
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from upath.types import JoinablePathLike


class DocumentConverter(ABC):
    """Base class for document converters."""

    def convert_file(self, path: JoinablePathLike) -> str:
        """Convert document file to markdown."""
        from upathtools import to_upath

        path_obj = to_upath(path)
        content = path_obj.read_bytes()
        return self.convert_content(content, mimetypes.guess_type(str(path))[0])

    @abstractmethod
    def convert_content(self, content: Any, mime_type: str | None = None) -> str:
        """Convert content to markdown."""

    @abstractmethod
    def supports_file(self, path: JoinablePathLike) -> bool:
        """Check if converter can handle this file type."""

    @abstractmethod
    def supports_content(self, content: Any, mime_type: str | None = None) -> bool:
        """Check if converter can handle this content type."""



"""Provider for resource access tools."""

from __future__ import annotations

from typing import TYPE_CHECKING

from llmling_agent.resource_providers.static import StaticResourceProvider
from llmling_agent.tools.base import Tool


if TYPE_CHECKING:
    from llmling import RuntimeConfig


def create_resource_access_tools(runtime: RuntimeConfig | None = None) -> list[Tool]:
    """Create tools for resource access operations."""
    tools: list[Tool] = []

    # Resource tools require runtime
    if runtime:
        tools.extend([
            Tool.from_callable(
                runtime.load_resource,
                source="builtin",
                category="read",
            ),
            Tool.from_callable(
                runtime.get_resources,
                source="builtin",
                category="search",
            ),
        ])

    return tools


class ResourceAccessTools(StaticResourceProvider):
    """Provider for resource access tools."""

    def __init__(
        self, name: str = "resource_access", runtime: RuntimeConfig | None = None
    ):
        super().__init__(name=name, tools=create_resource_access_tools(runtime))



"""Two-node team with a validator node."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from llmling_agent.delegation.teamrun import TeamRun


if TYPE_CHECKING:
    from llmling_agent import Agent, MessageNode


class StructuredTeam[TResult](TeamRun[Any, TResult]):
    """A team that produces typed/structured output through a processor node."""

    def __init__(
        self,
        worker: MessageNode[Any, Any],
        processor: Agent[Any, TResult],
        *,
        name: str | None = None,
        description: str | None = None,
    ):
        """Initialize structured team.

        Args:
            worker: The node doing the main work
            processor: Node that processes/validates the output
            name: Optional name for this team
            description: Optional description
        """
        super().__init__(
            agents=[worker, processor],
            name=name or f"{worker.name}>{processor.name}",
            description=description,
        )

    @property
    def worker(self) -> MessageNode[Any, Any]:
        """Get the worker node."""
        return self.agents[0]

    @worker.setter
    def worker(self, value: MessageNode[Any, Any]):
        """Set the worker node."""
        self.agents[0] = value

    @property
    def processor(self) -> Agent[Any, TResult]:
        """Get the processor node."""
        return self.agents[1]  # type: ignore

    @processor.setter
    def processor(self, value: Agent[Any, TResult]):
        """Set the processor node."""
        self.agents[1] = value



"""Event sources for LLMling agent."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from llmling_agent.log import get_logger


logger = get_logger(__name__)

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from llmling_agent.messaging.events import EventData
    from llmling_agent_config.events import EventConfig


class EventSource(ABC):
    """Base class for event sources."""

    @abstractmethod
    async def connect(self):
        """Initialize connection to event source."""

    @abstractmethod
    async def disconnect(self):
        """Close connection to event source."""

    @abstractmethod
    def events(self) -> AsyncGenerator[EventData]:
        """Get event iterator.

        Returns:
            AsyncIterator yielding events from this source

        Note: This is a coroutine that returns an AsyncIterator
        """
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: EventConfig) -> EventSource:
        """Create event source from configuration.

        Args:
            config: Event source configuration

        Returns:
            Configured event source instance

        Raises:
            ValueError: If source type is unknown or disabled
        """
        from llmling_agent_config.events import (
            EmailConfig,
            FileWatchConfig,
            TimeEventConfig,
            WebhookConfig,
        )

        if not config.enabled:
            msg = f"Source {config.name} is disabled"
            raise ValueError(msg)
        logger.info("Creating event source", name=config.name, provider=config.type)
        match config:
            case FileWatchConfig():
                from llmling_agent_events.file_watcher import FileSystemEventSource

                return FileSystemEventSource(config)
            case WebhookConfig():
                from llmling_agent_events.webhook_watcher import WebhookEventSource

                return WebhookEventSource(config)
            case EmailConfig():
                from llmling_agent_events.email_watcher import EmailEventSource

                return EmailEventSource(config)
            case TimeEventConfig():
                from llmling_agent_events.timed_watcher import TimeEventSource

                return TimeEventSource(config)
            case _:
                msg = f"Unknown event source type: {config.type}"
                raise ValueError(msg)



"""In-memory storage provider for testing."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, cast

from llmling_agent.messaging.messages import ChatMessage, TokenCost
from llmling_agent.storage import deserialize_messages
from llmling_agent.utils.now import get_now
from llmling_agent_storage.base import StorageProvider
from llmling_agent_storage.models import ConversationData


if TYPE_CHECKING:
    from collections.abc import Sequence

    from llmling_agent.common_types import JsonValue
    from llmling_agent_config.session import SessionQuery
    from llmling_agent_config.storage import MemoryStorageConfig
    from llmling_agent_storage.models import (
        MessageData,
        QueryFilters,
        StatsFilters,
        TokenUsage,
    )


class MemoryStorageProvider(StorageProvider):
    """In-memory storage provider for testing."""

    can_load_history = True

    def __init__(self, config: MemoryStorageConfig):
        super().__init__(config)
        self.messages: list[dict] = []
        self.conversations: list[dict] = []
        self.commands: list[dict] = []

    def cleanup(self):
        """Clear all stored data."""
        self.messages.clear()
        self.conversations.clear()
        self.commands.clear()

    async def filter_messages(self, query: SessionQuery) -> list[ChatMessage[str]]:
        """Filter messages from memory."""
        from llmling_agent.messaging.messages import ChatMessage

        filtered = []
        for msg in self.messages:
            # Skip if conversation ID doesn't match
            if query.name and msg["conversation_id"] != query.name:
                continue

            # Skip if agent name doesn't match
            if query.agents and not (
                msg["name"] in query.agents
                or (
                    query.include_forwarded
                    and msg["forwarded_from"]
                    and any(a in query.agents for a in msg["forwarded_from"])
                )
            ):
                continue

            # Skip if before cutoff time
            if query.since and (cutoff := query.get_time_cutoff()):  # noqa: SIM102
                if msg["timestamp"] < cutoff:
                    continue

            # Skip if after until time
            if query.until and msg["timestamp"] > datetime.fromisoformat(query.until):
                continue

            # Skip if content doesn't match search
            if query.contains and query.contains not in msg["content"]:
                continue

            # Skip if role doesn't match
            if query.roles and msg["role"] not in query.roles:
                continue

            # Convert cost info
            cost_info = None
            if msg["cost_info"]:
                total = msg.get("cost", 0.0)
                cost_info = TokenCost(token_usage=msg["cost_info"], total_cost=total)

            # Create ChatMessage
            chat_message = ChatMessage(
                content=msg["content"],
                role=msg["role"],
                name=msg["name"],
                model_name=msg["model"],
                cost_info=cost_info,
                response_time=msg["response_time"],
                forwarded_from=msg["forwarded_from"] or [],
                timestamp=msg["timestamp"],
                provider_name=msg["provider_name"],
                provider_response_id=msg["provider_response_id"],
                messages=deserialize_messages(msg["messages"]),
                finish_reason=msg["finish_reason"],
            )
            filtered.append(chat_message)

            # Apply limit if specified
            if query.limit and len(filtered) >= query.limit:
                break

        return filtered

    async def log_message(
        self,
        *,
        conversation_id: str,
        message_id: str,
        content: str,
        role: str,
        name: str | None = None,
        cost_info: TokenCost | None = None,
        model: str | None = None,
        response_time: float | None = None,
        forwarded_from: list[str] | None = None,
        provider_name: str | None = None,
        provider_response_id: str | None = None,
        messages: str | None = None,
        finish_reason: str | None = None,
    ):
        """Store message in memory."""
        if next((i for i in self.messages if i["message_id"] == message_id), None):
            msg = f"Duplicate message ID: {message_id}"
            raise ValueError(msg)

        self.messages.append({
            "conversation_id": conversation_id,
            "message_id": message_id,
            "content": content,
            "role": role,
            "name": name,
            "cost_info": cost_info.token_usage if cost_info else None,
            "model": model,
            "response_time": response_time,
            "forwarded_from": forwarded_from,
            "provider_name": provider_name,
            "provider_response_id": provider_response_id,
            "messages": messages,
            "finish_reason": finish_reason,
            "timestamp": get_now(),
        })

    async def log_conversation(
        self,
        *,
        conversation_id: str,
        node_name: str,
        start_time: datetime | None = None,
    ):
        """Store conversation in memory."""
        if next(
            (i for i in self.messages if i["conversation_id"] == conversation_id), None
        ):
            msg = f"Duplicate conversation ID: {conversation_id}"
            raise ValueError(msg)
        self.conversations.append({
            "id": conversation_id,
            "agent_name": node_name,
            "start_time": start_time or get_now(),
        })

    async def log_command(
        self,
        *,
        agent_name: str,
        session_id: str,
        command: str,
        context_type: type | None = None,
        metadata: dict[str, JsonValue] | None = None,
    ):
        """Store command in memory."""
        self.commands.append({
            "agent_name": agent_name,
            "session_id": session_id,
            "command": command,
            "timestamp": get_now(),
            "context_type": context_type.__name__ if context_type else None,
            "metadata": metadata or {},
        })

    async def get_commands(
        self,
        agent_name: str,
        session_id: str,
        *,
        limit: int | None = None,
        current_session_only: bool = False,
    ) -> list[str]:
        """Get commands from memory."""
        filtered = []
        for cmd in reversed(self.commands):  # newest first
            if current_session_only and cmd["session_id"] != session_id:
                continue
            if not current_session_only and cmd["agent_name"] != agent_name:
                continue
            filtered.append(cmd["command"])
            if limit and len(filtered) >= limit:
                break
        return filtered

    async def get_conversations(
        self,
        filters: QueryFilters,
    ) -> list[tuple[ConversationData, Sequence[ChatMessage[str]]]]:
        """Get filtered conversations from memory."""
        results: list[tuple[ConversationData, Sequence[ChatMessage[str]]]] = []

        # First get matching conversations
        convs = {}
        for conv in self.conversations:
            if filters.agent_name and conv["agent_name"] != filters.agent_name:
                continue
            if filters.since and conv["start_time"] < filters.since:
                continue
            convs[conv["id"]] = conv

        # Then get messages for each conversation
        for conv_id, conv in convs.items():
            conv_messages: list[ChatMessage[str]] = []
            for msg in self.messages:
                if msg["conversation_id"] != conv_id:
                    continue
                if filters.query and filters.query not in msg["content"]:
                    continue
                if filters.model and msg["model"] != filters.model:
                    continue

                cost_info = None
                if msg["cost_info"]:
                    total = msg.get("cost", 0.0)
                    cost_info = TokenCost(token_usage=msg["cost_info"], total_cost=total)

                chat_msg = ChatMessage[str](
                    content=msg["content"],
                    role=msg["role"],
                    name=msg["name"],
                    model_name=msg["model"],
                    cost_info=cost_info,
                    response_time=msg["response_time"],
                    forwarded_from=msg["forwarded_from"],
                    timestamp=msg["timestamp"],
                )
                conv_messages.append(chat_msg)

            # Skip if no matching messages for content filter
            if filters.query and not conv_messages:
                continue

            # Convert ChatMessages to MessageData format for ConversationData
            message_data: list[MessageData] = [
                cast(
                    "MessageData",
                    {
                        "role": msg.role,
                        "content": msg.content,
                        "timestamp": msg.timestamp.isoformat(),
                        "model": msg.model_name,
                        "name": msg.name,
                        "token_usage": msg.cost_info.token_usage
                        if msg.cost_info
                        else None,
                        "cost": msg.cost_info.total_cost if msg.cost_info else None,
                        "response_time": msg.response_time,
                    },
                )
                for msg in conv_messages
            ]

            # Create conversation data with proper MessageData
            conv_data = ConversationData(
                id=conv_id,
                agent=conv["agent_name"],
                start_time=conv["start_time"].isoformat(),
                messages=message_data,  # Now using properly typed MessageData
                token_usage=self._aggregate_token_usage(conv_messages),
            )
            results.append((conv_data, conv_messages))

            if filters.limit and len(results) >= filters.limit:
                break

        return results

    async def get_conversation_stats(
        self,
        filters: StatsFilters,
    ) -> dict[str, dict[str, Any]]:
        """Get statistics from memory."""
        # Collect raw data
        rows = []
        for msg in self.messages:
            if msg["timestamp"] <= filters.cutoff:
                continue
            if filters.agent_name and msg["name"] != filters.agent_name:
                continue

            cost_info = None
            if msg["cost_info"]:
                total = msg.get("cost", 0.0)
                cost_info = TokenCost(token_usage=msg["cost_info"], total_cost=total)

            rows.append((msg["model"], msg["name"], msg["timestamp"], cost_info))

        # Use base class aggregation
        return self.aggregate_stats(rows, filters.group_by)

    @staticmethod
    def _aggregate_token_usage(messages: Sequence[ChatMessage[Any]]) -> TokenUsage:
        """Sum up tokens from a sequence of messages."""
        total = prompt = completion = 0
        for msg in messages:
            if msg.cost_info:
                total += msg.cost_info.token_usage.total_tokens
                prompt += msg.cost_info.token_usage.input_tokens
                completion += msg.cost_info.token_usage.output_tokens
        return {"total": total, "prompt": prompt, "completion": completion}

    async def reset(
        self,
        *,
        agent_name: str | None = None,
        hard: bool = False,
    ) -> tuple[int, int]:
        """Reset stored data."""
        # Get counts first
        conv_count, msg_count = await self.get_conversation_counts(agent_name=agent_name)

        if hard:
            if agent_name:
                msg = "Hard reset cannot be used with agent_name"
                raise ValueError(msg)
            # Clear everything
            self.cleanup()
            return conv_count, msg_count

        if agent_name:
            # Filter out data for specific agent
            self.conversations = [
                c for c in self.conversations if c["agent_name"] != agent_name
            ]
            self.messages = [
                m
                for m in self.messages
                if m["conversation_id"]
                not in {
                    c["id"] for c in self.conversations if c["agent_name"] == agent_name
                }
            ]
        else:
            # Clear all
            self.messages.clear()
            self.conversations.clear()
            self.commands.clear()

        return conv_count, msg_count

    async def get_conversation_counts(
        self,
        *,
        agent_name: str | None = None,
    ) -> tuple[int, int]:
        """Get conversation and message counts."""
        if agent_name:
            conv_count = sum(
                1 for c in self.conversations if c["agent_name"] == agent_name
            )
            msg_count = sum(
                1
                for m in self.messages
                if any(
                    c["id"] == m["conversation_id"] and c["agent_name"] == agent_name
                    for c in self.conversations
                )
            )
        else:
            conv_count = len(self.conversations)
            msg_count = len(self.messages)

        return conv_count, msg_count



# /// script
# dependencies = ["llmling-agent"]
# ///

"""Demo: Agent using MCP server with code fixer (sampling + elicitation)."""

from __future__ import annotations

import asyncio
from pathlib import Path

from llmling_agent import Agent
from llmling_agent_config.mcp_server import StdioMCPServerConfig


async def main():
    """Demo MCP server with code fixer workflow."""
    print("🚀 Starting code fixer demo...")

    # Get server path
    server_path = Path(__file__).parent / "server.py"

    # Create MCP server config
    mcp_server = StdioMCPServerConfig(
        name="code_fixer_demo",
        command="uv",
        args=["run", str(server_path)],
    )

    # Create agent with MCP server
    agent = Agent(
        name="demo_agent",
        model="openai:gpt-5-nano",
        system_prompt="You are a helpful assistant with code fixing tools.",
        mcp_servers=[mcp_server],
    )

    async with agent:
        # Code with actual bugs
        buggy_code = 'prin("hello world"'

        print("\n" + "=" * 60)
        print("Demo: Code Fixer (Sampling + Elicitation)")
        print(f"Original code: {buggy_code}")
        print("=" * 60)

        result = await agent.run(
            f"Please use fix_code to analyze and fix this code: {buggy_code}"
        )
        print(f"\n✅ Agent response:\n{result.data}")

        print("\n✨ Code fixer demo completed!")


if __name__ == "__main__":
    asyncio.run(main())



"""Agent delegation and collaboration functionality."""

from llmling_agent.delegation.pool import AgentPool
from llmling_agent.delegation.base_team import BaseTeam
from llmling_agent.delegation.teamrun import TeamRun
from llmling_agent.delegation.team import Team

__all__ = ["AgentPool", "BaseTeam", "Team", "TeamRun"]



"""Utilities for creating Typer commands from Pydantic models."""

from __future__ import annotations

import inspect
import types as pytypes
from typing import TYPE_CHECKING, Annotated, Any, Union, get_args, get_origin

import typer


if TYPE_CHECKING:
    from collections.abc import Callable
    from inspect import Parameter

    from pydantic import BaseModel


def resolve_type(field_type: Any) -> Any:
    """Resolve actual type from Union/Optional types."""
    origin = get_origin(field_type)

    # Handle Any
    if field_type is Any:
        return str

    # Handle both Union and | syntax
    if origin is Union or origin is pytypes.UnionType:
        types_ = [t for t in get_args(field_type) if t is not type(None)]
        return resolve_type(types_[0]) if len(types_) == 1 else str

    # Handle lists/sets
    if origin in {list, set}:
        item_type = get_args(field_type)[0]
        return list[resolve_type(item_type)]  # type: ignore

    return field_type


def create_typer_command[T: BaseModel](
    config_cls: type[T],
    callback: Callable[[T], Any] | None = None,
    *,
    name: str | None = None,
) -> typer.models.CommandInfo:
    """Create a Typer CommandInfo from a Pydantic model."""

    def make_command(**kwargs: Any) -> Any:
        """Command created from model."""
        config = config_cls(**kwargs)
        if callback:
            return callback(config)
        return config

    # Create parameters from model fields
    params: dict[str, Parameter] = {}
    for field_name, field in config_cls.model_fields.items():
        if field_name == "type":  # Skip discriminator
            continue

        field_type = resolve_type(field.annotation)
        default = field.default if field.default != ... else None

        # First field becomes argument
        if not params:
            # For arguments, the param name should be the field name
            params[field_name] = inspect.Parameter(
                field_name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Annotated[field_type, typer.Argument(help=field.description)],
                default=default,
            )
            continue

        # For options, we need proper param_decls
        option_name = field_name.replace("_", "-")
        params[field_name] = inspect.Parameter(
            field_name,
            inspect.Parameter.KEYWORD_ONLY,
            annotation=Annotated[
                field_type,
                typer.Option(
                    "--" + option_name,  # Explicit param_decls
                    help=field.description,
                ),
            ],
            default=default,
        )

    # Create function with signature
    make_command.__signature__ = inspect.Signature(  # type: ignore
        list(params.values()), return_annotation=config_cls
    )
    make_command.__annotations__ = {k: v.annotation for k, v in params.items()}

    # Create command
    if name is None:
        name = config_cls.__name__.lower().replace("command", "")

    return typer.models.CommandInfo(
        name=name,
        callback=make_command,
        help=config_cls.__doc__,
    )



"""Provider for RuntimeConfig tools."""

from __future__ import annotations

from typing import TYPE_CHECKING

from llmling_agent.log import get_logger
from llmling_agent.resource_providers.base import ResourceProvider
from llmling_agent.tools.base import Tool
from llmling_agent_config.resources import ResourceInfo


if TYPE_CHECKING:
    from llmling import RuntimeConfig
    from llmling.prompts import BasePrompt

logger = get_logger(__name__)


class RuntimeResourceProvider(ResourceProvider):
    """Provider that exposes RuntimeConfig tools through ResourceProvider interface."""

    def __init__(self, runtime: RuntimeConfig, name: str = "runtime"):
        """Initialize provider with RuntimeConfig.

        Args:
            runtime: RuntimeConfig instance to wrap
            name: Name of the provider
        """
        super().__init__(name=name)
        self._runtime = runtime

    async def get_tools(self) -> list[Tool]:
        """Convert RuntimeConfig tools to Tools."""
        tools: list[Tool] = []

        for tool in self._runtime.get_tools():
            try:
                tools.append(Tool(tool, source="runtime"))
            except Exception:
                logger.exception("Failed to convert runtime tool", name=tool.name)
                continue

        return tools

    async def get_prompts(self) -> list[BasePrompt]:
        """Get prompts from runtime."""
        return list(self._runtime.get_prompts())

    async def get_resources(self) -> list[ResourceInfo]:
        """Convert runtime resources to ResourceInfo."""
        resources: list[ResourceInfo] = []

        for resource in self._runtime.get_resources():
            try:
                name = resource.name or str(resource.uri)
                uri = str(resource.uri)
                info = ResourceInfo(name=name, uri=uri, description=resource.description)
                resources.append(info)
            except Exception:
                logger.exception("Failed to convert runtime resource", name=resource.name)
                continue

        return resources



"""Client ACP Connection."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol


if TYPE_CHECKING:
    from acp.schema import (
        AuthenticateRequest,
        AuthenticateResponse,
        CancelNotification,
        InitializeRequest,
        InitializeResponse,
        LoadSessionRequest,
        LoadSessionResponse,
        NewSessionRequest,
        NewSessionResponse,
        PromptRequest,
        PromptResponse,
        SetSessionModelRequest,
        SetSessionModelResponse,
        SetSessionModeRequest,
        SetSessionModeResponse,
    )


class BaseAgent(Protocol):
    """Base agent interface for ACP - always required."""

    async def initialize(self, params: InitializeRequest) -> InitializeResponse: ...

    async def new_session(self, params: NewSessionRequest) -> NewSessionResponse: ...

    async def prompt(self, params: PromptRequest) -> PromptResponse: ...

    async def cancel(self, params: CancelNotification) -> None: ...


class SessionPersistenceCapability(Protocol):
    """Agent capability for session persistence and authentication."""

    async def load_session(self, params: LoadSessionRequest) -> LoadSessionResponse: ...

    async def authenticate(
        self, params: AuthenticateRequest
    ) -> AuthenticateResponse | None: ...


class SessionModeCapability(Protocol):
    """Agent capability for session mode switching."""

    async def set_session_mode(
        self, params: SetSessionModeRequest
    ) -> SetSessionModeResponse | None: ...


class SessionModelCapability(Protocol):
    """Agent capability for session model switching."""

    async def set_session_model(
        self, params: SetSessionModelRequest
    ) -> SetSessionModelResponse | None: ...


class AgentExtensibilityCapability(Protocol):
    """Agent capability for extension methods."""

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]: ...

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None: ...


class Agent(
    BaseAgent,
    SessionPersistenceCapability,
    SessionModeCapability,
    SessionModelCapability,
    AgentExtensibilityCapability,
):
    """ACP Agent interface.

    Backward compatibility class that includes all agent capabilities.
    New implementations should inherit from specific capability protocols instead.
    """



"""Session state schema definitions."""

from __future__ import annotations

from collections.abc import Sequence  # noqa: TC003

from acp.schema.base import AnnotatedObject


class ModelInfo(AnnotatedObject):
    """**UNSTABLE**.

    This capability is not part of the spec yet.

    Information about a selectable model.
    """

    description: str | None = None
    """Optional description of the model."""

    model_id: str
    """Unique identifier for the model."""

    name: str
    """Human-readable name of the model."""


class SessionModelState(AnnotatedObject):
    """**UNSTABLE**.

    This capability is not part of the spec yet.

    The set of models and the one currently active.
    """

    available_models: Sequence[ModelInfo]
    """The set of models that the Agent can use"""

    current_model_id: str
    """The current model the Agent is in."""


class SessionMode(AnnotatedObject):
    """A mode the agent can operate in.

    See protocol docs: [Session Modes](https://agentclientprotocol.com/protocol/session-modes)
    """

    description: str | None = None
    """Optional description of the mode."""

    id: str
    """Unique identifier for a Session Mode."""

    name: str
    """Human-readable name of the mode."""


class SessionModeState(AnnotatedObject):
    """The set of modes and the one currently active."""

    available_modes: Sequence[SessionMode]
    """The set of modes that the Agent can operate in"""

    current_mode_id: str
    """The current mode the Agent is in."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import fsspec
from fsspec import AbstractFileSystem
from llmling import BaseRegistry, LLMLingError
from upath import UPath
from upathtools import UnionFileSystem

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from llmling_agent_config.resources import ResourceConfig

logger = get_logger(__name__)


class ResourceRegistry(BaseRegistry[str, AbstractFileSystem]):
    """Registry for filesystem resources."""

    @property
    def _error_class(self) -> type[LLMLingError]:
        return LLMLingError

    def register(self, name: str, item: Any, replace: bool = False):
        """Register a new resource."""
        logger.debug("registering %r (%r)", name, item.__class__.__name__)
        fsspec.register_implementation(name, item.__class__, clobber=True)
        super().register(name, item, replace=replace)

    def _validate_item(self, item: Any) -> AbstractFileSystem:
        if not isinstance(item, AbstractFileSystem):
            msg = f"Expected AbstractFileSystem, got {type(item)}"
            raise self._error_class(msg)
        return item

    def register_from_config(
        self, name: str, config: ResourceConfig
    ) -> AbstractFileSystem:
        """Register a new resource from config."""
        from llmling_agent_config.resources import SourceResourceConfig

        match config:
            case SourceResourceConfig(uri=uri):
                # Extract base path from URI if present
                protocol, path = uri.split("://", 1) if "://" in uri else (uri, "")
                if path:
                    config.storage_options["root"] = path

                fs = fsspec.filesystem(protocol, **config.storage_options)

                if config.path:
                    fs = fsspec.filesystem("dir", fs=fs, path=config.path)
                if config.cached:
                    fs = fsspec.filesystem("cached", fs=fs)

                self.register(name, fs)
                return fs
            case _:
                msg = f"Unknown resource config type: {type(config)}"
                raise ValueError(msg)

    def get_fs(self) -> UnionFileSystem:
        """Get unified filesystem view of all resources."""
        filesystems = dict(self.items())
        return UnionFileSystem(filesystems)

    def get_upath(self, resource_name: str | None = None) -> UPath:
        """Get a UPath object for accessing a resource."""
        return UPath("union://")
        # path._fs_cached = self.get_fs()
        # return path

    async def get_content(
        self,
        path: str,
        encoding: str = "utf-8",
        recursive: bool = True,
        exclude: list[str] | None = None,
        max_depth: int | None = None,
    ) -> str:
        """Get content from a resource as text.

        Args:
            path: Path to read, either:
                - resource (whole resource)
                - resource:// (whole resource)
                - resource://file.txt (single file)
                - resource://folder (directory)
            encoding: Text encoding for binary content
            recursive: For directories, whether to read recursively
            exclude: For directories, patterns to exclude
            max_depth: For directories, maximum depth to read

        Returns:
            For files: file content
            For directories: concatenated content of all files
        """
        from upathtools import read_folder, read_path

        if "/" not in path:
            path = f"{path}://"
        _resource, _ = path.split("://", 1)
        fs = self.get_fs()
        is_dir = await fs._isdir(path)

        if is_dir:
            content_dict = await read_folder(
                path,
                mode="rt",
                encoding=encoding,
                recursive=recursive,
                exclude=exclude,
                max_depth=max_depth,
            )
            # Combine all files with headers
            sections = []
            for rel_path, content in sorted(content_dict.items()):
                sections.extend([f"--- {rel_path} ---", content, ""])
            return "\n".join(sections)
        return await read_path(path, encoding=encoding)

    async def query(
        self,
        path: str,
        pattern: str = "**/*",
        *,
        recursive: bool = True,
        include_dirs: bool = False,
        exclude: list[str] | None = None,
        max_depth: int | None = None,
    ) -> list[str]:
        """Query contents of a resource or subfolder.

        Args:
            path: Path to query, either:
                - resource (queries whole resource)
                - resource:// (queries whole resource)
                - resource://subfolder (queries specific folder)
            pattern: Glob pattern to match files against
            recursive: Whether to search subdirectories
            include_dirs: Whether to include directories in results
            exclude: List of patterns to exclude
            max_depth: Maximum directory depth for recursive search

        Example:
            # Query whole resource (all equivalent)
            files = await registry.query("docs")
            files = await registry.query("docs://")

            # Query specific subfolder
            files = await registry.query("docs://guides", pattern="*.md")
        """
        from upathtools import list_files

        if "/" not in path:
            # Simple resource name - add protocol
            path = f"{path}://"
        resource, _ = path.split("://", 1)
        if resource not in self:
            msg = f"Resource not found: {resource}"
            raise ValueError(msg)

        files = await list_files(
            path,
            pattern=pattern,
            recursive=recursive,
            include_dirs=include_dirs,
            exclude=exclude,
            max_depth=max_depth,
        )
        return [str(p) for p in files]


if __name__ == "__main__":
    registry = ResourceRegistry()
    p = registry.get_upath()
    print(p, type(p))
